from __future__ import annotations

import os
import pickle
import signal
import sys
import typing as ty
import urllib.parse
from collections import defaultdict
from contextlib import suppress
from pathlib import Path
from time import time

from gi.repository import Gdk, Gio, GLib, Gtk

try:
    from gi.repository import Wnck

    Wnck.set_client_type(Wnck.ClientType.PAGER)
    if "WAYLAND_DISPLAY" in os.environ:
        Wnck.Screen.get_default = lambda *_x: None

except ImportError as e:
    from kupfer.support import pretty

    pretty.print_info(__name__, "Disabling window tracking:", e)
    Wnck = None

from kupfer import config, desktop_launch
from kupfer.core import settings

## NOTE: SpawnError  *should* be imported from this module TODO: check
# pylint: disable=unused-import
from kupfer.desktop_launch import SpawnError
from kupfer.support import fileutils, pretty, scheduler, system
from kupfer.ui import uievents

__all__ = (
    "AsyncCommand",
    "SpawnError",
    "application_close_all",
    "application_id",
    "application_is_running",
    "get_applications_matcher_service",
    "get_display_path_for_bytestring",
    "launch_application",
    "show_help_url",
    "show_url",
    "spawn_async",
    "spawn_async_notify_as",
    "spawn_async_raise",
    "spawn_in_terminal",
    "spawn_terminal",
)

_DEFAULT_ASSOCIATIONS = {
    "evince": "Document Viewer",
    "file-roller": "File Roller",
    # "gedit" : "Text Editor",
    "gnome-keyring-manager": "Keyring Manager",
    "nautilus-browser": "File Manager",
    "rhythmbox": "Rhythmbox Music Player",
}


def application_id(
    app_info: Gio.AppInfo, desktop_file: str | None = None
) -> str:
    """Return an application id (string) for GAppInfo @app_info"""
    app_id = app_info.get_id() or desktop_file or ""
    return app_id.removesuffix(".desktop")


# pylint: disable=too-many-arguments
def launch_application(
    app_info: Gio.AppInfo,
    files: ty.Iterable[Gio.File] = (),
    uris: ty.Iterable[str] = (),
    paths: ty.Iterable[str] = (),
    track: bool = True,
    activate: bool = True,
    desktop_file: str | None = None,
    screen: Gdk.Screen | None = None,
    work_dir: str | None = None,
) -> bool:
    """Launch `app_rec` correctly, using a startup notification. You may pass
    inlist of `Gio.Files` in `files` and/or a list of `uris` and/or list of
    `paths`. This lists are combined together.

    if `track`, it is a user-level application.
    if `activate`, activate rather than start a new version.

    `app_rec` is either an `GAppInfo` or (`GAppInfo`, desktop_file_path) tuple.

    `work_dir`: overwrite work directory of application.

    Raises `SpawnError` on failed program start.
    """
    assert app_info
    assert not (bool(paths) and bool(uris)), (
        "either paths or uris must be given: " + repr((paths, uris))
    )

    svc = get_applications_matcher_service()
    app_id = application_id(app_info, desktop_file)

    if activate and svc.application_is_running(app_id):
        svc.application_to_front(app_id)
        return True

    launch_callback = None
    if track:
        # An launch callback closure for the @app_id
        def app_launch_callback(argv, pid, _notify_id, _files, _timestamp):
            if not settings.is_known_terminal_executable(argv[0]):
                svc.launched_application(app_id, pid)

        launch_callback = app_launch_callback

    files = list(files)
    if paths:
        files.extend(map(Gio.File.new_for_path, paths))

    if uris:
        files.extend(map(Gio.File.new_for_uri, uris))

    desktop_launch.launch_app_info(
        app_info,
        files,
        timestamp=uievents.current_event_time(),
        desktop_file=desktop_file,
        launch_cb=launch_callback,
        screen=screen,
        work_dir=work_dir,
    )
    return True


def application_is_running(app_id: str) -> bool:
    svc = get_applications_matcher_service()
    return svc.application_is_running(app_id)


def application_close_all(app_id: str) -> None:
    svc = get_applications_matcher_service()
    svc.application_close_all(app_id)


class ApplicationsMatcherService(pretty.OutputMixin):
    """Handle launching applications and see if they still run.
    This is a learning service, since we have no first-class application
    object on the Linux desktop
    """

    _instance: ApplicationsMatcherService | None = None

    @classmethod
    def instance(cls) -> ApplicationsMatcherService:
        if cls._instance is None:
            cls._instance = ApplicationsMatcherService()

        return cls._instance

    def __init__(self):
        self.register: dict[str, str] = {}
        self._get_wnck_screen_windows_stacked()
        scheduler.get_scheduler().connect("finish", self._on_finish)
        self._load()

    @classmethod
    def _get_wnck_screen_windows_stacked(cls) -> ty.Iterable["Wnck.Window"]:
        if Wnck and (screen := Wnck.Screen.get_default()):
            return screen.get_windows_stacked()  # type: ignore

        return ()

    def _get_filename(self) -> str:
        # Version 1: Up to incl v203
        # Version 2: Do not track terminals
        version = 2
        return os.path.join(
            config.get_cache_home() or "",
            f"application_identification_v{version}.pickle",
        )

    def _load(self) -> None:
        reg = self._unpickle_register(self._get_filename())
        self.register = reg or _DEFAULT_ASSOCIATIONS
        # pretty-print register to debug
        if self.register:
            self.output_debug("Learned the following applications")
            items = "\n".join(
                f"  {k:<30} : {v}" for k, v in self.register.items()
            )
            self.output_debug(f"\n{{\n{items}\n}}")

    def _on_finish(self, _sched: ty.Any) -> None:
        self._pickle_register(self.register, self._get_filename())

    def _unpickle_register(self, pickle_file: str) -> dict[str, str] | None:
        try:
            source = pickle.loads(Path(pickle_file).read_bytes())
            assert isinstance(source, dict), "Stored object not a dict"
            self.output_debug(f"Reading from {pickle_file}")
            return source
        except OSError:
            pass
        except (pickle.PickleError, Exception) as exc:
            self.output_info(f"Error loading {pickle_file}: {exc}")

        return None

    def _pickle_register(self, reg: ty.Any, pickle_file: str) -> bool:
        self.output_debug(f"Saving to {pickle_file}")
        Path(pickle_file).write_bytes(
            pickle.dumps(reg, pickle.HIGHEST_PROTOCOL)
        )
        return True

    def _store(self, app_id: str, window: "Wnck.Window") -> None:
        # FIXME: Store the 'res_class' name?
        application = window.get_application()
        res_class = window.get_class_group().get_res_class()
        store_name = application.get_name()
        self.register[app_id] = res_class
        self.output_debug(
            "storing application",
            app_id,
            "as",
            store_name,
            "res_class",
            res_class,
        )

    def _has_match(self, app_id: str | None) -> bool:
        if not app_id:
            return False

        return app_id in self.register

    def _is_match(self, app_id: str, window: "Wnck.Window") -> bool:
        application = window.get_application()
        res_class = window.get_class_group().get_res_class()
        reg_name = self.register.get(app_id)
        if reg_name and reg_name in (application.get_name(), res_class):
            return True

        return app_id in (application.get_name().lower(), res_class.lower())

    def launched_application(self, app_id: str, pid: int) -> None:
        if self._has_match(app_id):
            return

        timeout = time() + 15
        GLib.timeout_add_seconds(
            2, self._find_application, app_id, pid, timeout
        )
        # and once later
        GLib.timeout_add_seconds(
            30, self._find_application, app_id, pid, timeout
        )

    def _find_application(self, app_id: str, pid: int, timeout: float) -> bool:
        if self._has_match(app_id):
            return False

        # self.output_debug("Looking for window for application", app_id)
        for win in self._get_wnck_screen_windows_stacked():
            app = win.get_application()
            app_pid = app.get_pid() or win.get_pid()
            if app_pid == pid:
                self._store(app_id, win)
                return False

        return time() <= timeout

    def application_name(self, app_id: str | None) -> str | None:
        if not app_id:
            return None

        return self.register.get(app_id)

    def application_is_running(self, app_id: str) -> bool:
        for win in self._get_wnck_screen_windows_stacked():
            if (
                win.get_application()
                and self._is_match(app_id, win)
                and win.get_window_type() == Wnck.WindowType.NORMAL
            ):
                return True

        return False

    def _get_application_windows(
        self, app_id: str
    ) -> ty.Iterator["Wnck.Window"]:
        if not Wnck:
            return

        for win in self._get_wnck_screen_windows_stacked():
            if (
                win.get_application()
                and self._is_match(app_id, win)
                and win.get_window_type() == Wnck.WindowType.NORMAL
            ):
                yield win

    def application_to_front(self, app_id: str) -> None:
        application_windows = list(self._get_application_windows(app_id))
        if not application_windows:
            return

        etime = uievents.current_event_time()
        # if True, focus app's all windows on the same workspace
        # if False, focus only one window (in cyclical manner)
        focus_all = True
        if focus_all:
            self._to_front_application_style(application_windows, etime)
            return

        self._to_front_single(application_windows, etime)

    def _to_front_application_style(
        self, application_windows: list["Wnck.Window"], evttime: int
    ) -> None:
        workspaces: dict[Wnck.Workspace, list[Wnck.Window]] = defaultdict(list)
        cur_screen = application_windows[0].get_screen()
        cur_workspace = cur_screen.get_active_workspace()

        ## get all visible windows in stacking order
        vis_windows = [
            win
            for win in self._get_wnck_screen_windows_stacked()
            if (
                win.get_window_type() == Wnck.WindowType.NORMAL
                and win.is_visible_on_workspace(cur_workspace)
            )
        ]

        ## sort windows into "bins" by workspace
        for win in application_windows:
            if win.get_window_type() == Wnck.WindowType.NORMAL:
                wspc = win.get_workspace() or cur_workspace
                workspaces[wspc].append(win)

        cur_wspc_windows = workspaces[cur_workspace]
        # make a rotated workspace list, with current workspace first
        all_workspaces = cur_screen.get_workspaces()
        all_workspaces.pop(cur_workspace.get_number())
        # check if the application's window on current workspace
        # are the topmost
        focus_windows = []
        if cur_wspc_windows and set(
            vis_windows[-len(cur_wspc_windows) :]
        ) != set(cur_wspc_windows):
            focus_windows = cur_wspc_windows
            ## if the topmost window is already active, take another
            if focus_windows[-1] == vis_windows[-1]:
                focus_windows.pop()
        else:
            # all windows are focused, find on next workspace
            for wspc in all_workspaces[1:]:
                if focus_windows := workspaces[wspc]:
                    break
            else:
                # no windows on other workspaces, so we rotate among
                # the local ones
                focus_windows = cur_wspc_windows[:1]

        self._focus_windows(focus_windows, evttime)

    def _to_front_single(
        self, application_windows: list["Wnck.Window"], evttime: int
    ) -> None:
        # bring the first window to front
        for window in application_windows:
            self._focus_windows([window], evttime)
            return

    def _focus_windows(
        self, windows: list["Wnck.Window"], evttime: int
    ) -> None:
        for window in windows:
            # we special-case the desktop
            # only show desktop if it's the only window
            if window.get_name() == "x-nautilus-desktop":
                if len(windows) == 1:
                    if screen := Wnck.Screen.get_default():
                        screen.toggle_showing_desktop(True)
                else:
                    continue

            if wspc := window.get_workspace():
                wspc.activate(evttime)

            window.activate_transient(evttime)

    def application_close_all(self, app_id: str) -> None:
        application_windows = self._get_application_windows(app_id)
        evttime = uievents.current_event_time()
        for win in application_windows:
            if not win.is_skip_tasklist():
                win.close(evttime)


# Get the (singleton) ApplicationsMatcherService"
get_applications_matcher_service = ApplicationsMatcherService.instance


def _split_string(inp: bytes, length: int) -> ty.Iterator[bytes]:
    """Split @inp in pieces of @length

    >>> list(_split_string(b"abcdefghijk", 3))
    [b"abc", b"def", b"ghi", b"jk"]

    """
    while inp:
        yield inp[:length]
        inp = inp[length:]


class AsyncCommand(pretty.OutputMixin):
    """Run a command asynchronously (using the GLib mainloop).

    call @finish_callback when command terminates, or
    when command is killed after @timeout_s seconds, whichever
    comes first.

    If @timeout_s is None, no timeout is used

    If stdin is a byte string, it is supplied on the command's stdin.

    If env is None, command will inherit the parent's environment.

    finish_callback -> (AsyncCommand, stdout_output, stderr_output)

    Attributes:
    self.exit_status  Set after process exited
    self.finished     bool
    """

    # the maximum input (bytes) we'll read in one shot (one io_callback)
    max_input_buf = 512 * 1024

    def __init__(
        self,
        argv: list[str],
        finish_callback: ty.Callable[[AsyncCommand, bytes, bytes], None],
        timeout_s: int | None,
        stdin: bytes | None = None,
        env: ty.Any = None,
    ) -> None:
        self.stdout: list[bytes] = []
        self.stderr: list[bytes] = []
        self.stdin: list[bytes] = []
        self.timeout = False
        self.killed = False
        self.finished = False
        self.finish_callback = finish_callback
        self.exit_status: int | None = None

        self.output_debug("AsyncCommand:", argv)

        flags = GLib.SPAWN_SEARCH_PATH | GLib.SPAWN_DO_NOT_REAP_CHILD
        kwargs = {}
        if env is not None:
            kwargs["envp"] = env

        pid, stdin_fd, stdout_fd, stderr_fd = GLib.spawn_async(
            argv,
            standard_output=True,
            standard_input=True,
            standard_error=True,
            flags=flags,
            **kwargs,
        )

        if stdin:
            self.stdin = list(_split_string(stdin, self.max_input_buf))
            in_io_flags = (
                GLib.IO_OUT | GLib.IO_ERR | GLib.IO_HUP | GLib.IO_NVAL
            )
            GLib.io_add_watch(
                stdin_fd, in_io_flags, self._in_io_callback, self.stdin
            )
        else:
            os.close(stdin_fd)

        io_flags = GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP | GLib.IO_NVAL
        GLib.io_add_watch(stdout_fd, io_flags, self._io_callback, self.stdout)
        GLib.io_add_watch(stderr_fd, io_flags, self._io_callback, self.stderr)
        self.pid = pid
        GLib.child_watch_add(pid, self._child_callback)
        if timeout_s is not None:
            GLib.timeout_add_seconds(timeout_s, self._timeout_callback)

    def _io_callback(
        self, sourcefd: int, condition: int, databuf: list[bytes]
    ) -> bool:
        if condition & GLib.IO_IN:
            databuf.append(os.read(sourcefd, self.max_input_buf))
            return True

        return False

    def _in_io_callback(
        self, sourcefd: int, condition: int, databuf: list[bytes]
    ) -> bool:
        """write to child's stdin"""
        if condition & GLib.IO_OUT:
            if not databuf:
                os.close(sourcefd)
                return False

            data = databuf.pop(0)
            written = os.write(sourcefd, data)
            if written < len(data):
                databuf.insert(0, data[written:])

            return True

        return False

    def _child_callback(self, pid: int, condition: int) -> None:
        # @condition is the &status field of waitpid(2) (C library)
        self.exit_status = os.WEXITSTATUS(condition)
        self.finished = True
        self.finish_callback(
            self, b"".join(self.stdout), b"".join(self.stderr)
        )

    def _timeout_callback(self) -> None:
        "send term signal on timeout"
        if not self.finished:
            self.timeout = True
            os.kill(self.pid, signal.SIGTERM)
            GLib.timeout_add_seconds(2, self._kill_callback)

    def _kill_callback(self) -> None:
        "Last resort, send kill signal"
        if not self.finished:
            self.killed = True
            os.kill(self.pid, signal.SIGKILL)


def spawn_terminal(
    workdir: str | None = None, screen: str | None = None
) -> bool:
    "Raises SpawnError"
    term = settings.get_configured_terminal()
    if not term:
        return False

    notify = term["startup_notify"]
    app_id = term["desktopid"]
    argv = term["argv"]
    return desktop_launch.spawn_app_id(app_id, argv, workdir, notify, screen)


def spawn_in_terminal(argv: list[str], workdir: str | None = None) -> bool:
    "Raises SpawnError"
    term = settings.get_configured_terminal()
    if not term:
        return False

    notify = term["startup_notify"]
    _argv = list(term["argv"])
    if term["exearg"]:
        _argv.append(term["exearg"])

    _argv.extend(argv)
    return desktop_launch.spawn_app_id(
        term["desktopid"], _argv, workdir, notify
    )


def spawn_async_notify_as(app_id: str, argv: list[str]) -> bool:
    """
    Spawn argument list @argv and startup-notify as
    if application @app_id is starting (if possible)

    raises SpawnError
    """
    return desktop_launch.spawn_app_id(app_id, argv, None, True)


def spawn_async(
    argv: ty.Collection[str],
    in_dir: str = ".",
    *,
    finish_callback: ty.Callable[[int], None] | None = None,
) -> bool:
    """
    Silently spawn @argv in the background

    Returns False on failure
    """
    try:
        return spawn_async_raise(argv, in_dir, finish_callback=finish_callback)
    except SpawnError as exc:
        pretty.print_debug(__name__, "spawn_async", argv, exc)
        return False


def spawn_async_raise(
    argv: ty.Collection[str],
    workdir: str = ".",
    *,
    finish_callback: ty.Callable[[int], None] | None = None,
) -> bool:
    """
    A version of spawn_async that raises on error.

    raises SpawnError
    """
    pretty.print_debug(__name__, "spawn_async", argv, workdir)
    try:
        res = GLib.spawn_async(
            argv, working_directory=workdir, flags=GLib.SPAWN_SEARCH_PATH
        )
        if finish_callback and res:

            def _callback(pid: int, condition: int) -> None:
                finish_callback(os.WEXITSTATUS(condition))

            GLib.child_watch_add(res[0], _callback)

        return bool(res)
    except GLib.GError as exc:
        raise SpawnError(exc.message) from exc  # pylint: disable=no-member

    return False


def _try_register_pr_pdeathsig() -> None:
    """
    Register pr_set_pdeathsig (linux-only) for the calling process
    which is a signal delivered when its parent dies.

    This should ensure child processes die with the parent.
    """
    pr_set_pdeathsig = 1
    sighup = 1
    if sys.platform != "linux2":
        return

    with suppress(ImportError):
        # pylint: disable=import-outside-toplevel
        import ctypes

    with suppress(AttributeError, OSError):
        libc = ctypes.CDLL("libc.so.6")
        libc.prctl(pr_set_pdeathsig, sighup)


def _on_child_exit(
    pid: int, condition: int, user_data: tuple[ty.Any, bool]
) -> None:
    # @condition is the &status field of waitpid(2) (C library)
    argv, respawn = user_data
    if respawn:
        is_signal = os.WIFSIGNALED(condition)
        if is_signal and respawn:

            def callback(*args):
                _spawn_child(*args)
                return False

            GLib.timeout_add_seconds(10, callback, argv, respawn)


def _spawn_child(
    argv: list[str], respawn: bool = True, display: str | None = None
) -> int:
    """
    Spawn argv in the mainloop and keeping it as a child process
    (it will be made sure to exit with the parent).

    @respawn: If True, respawn if child dies abnormally

    raises launch.SpawnError
    returns pid
    """
    flags = GLib.SPAWN_SEARCH_PATH | GLib.SPAWN_DO_NOT_REAP_CHILD
    envp: list[str] = []
    if display:
        # environment is passed as a sequence of strings
        envd = os.environ.copy()
        envd["DISPLAY"] = display
        envp = ["=".join((k, v)) for k, v in envd.items()]

    try:
        pid: int
        pid, *_fds = GLib.spawn_async(
            argv,
            envp,
            flags=flags,
            child_setup=_try_register_pr_pdeathsig,
        )
    except GLib.GError as exc:
        raise SpawnError(str(exc)) from exc

    if pid:
        GLib.child_watch_add(pid, _on_child_exit, (argv, respawn))

    return pid


def start_plugin_helper(
    name: str, respawn: bool, display: str | None = None
) -> int:
    """
    @respawn: If True, respawn if child dies abnormally

    raises SpawnError

    UNUSED
    """
    argv = [sys.executable]
    argv.extend(sys.argv)
    argv.append(f"--exec-helper={name}")
    pretty.print_debug(__name__, "Spawning", argv)
    return _spawn_child(argv, respawn, display=display)


def show_path(path: str) -> None:
    """Open local @path with default viewer"""
    # Implemented using Gtk.show_uri
    gfile = Gio.File.new_for_path(path)
    if not gfile:
        return

    url = gfile.get_uri()
    show_url(url)


def show_url(url: str) -> bool:
    """Open any @url with default viewer"""

    # if there is no schema w url add default http://
    # this allow open urls like 'www.foobar.com'.
    url = url.strip()

    purl = urllib.parse.urlparse(url)
    if not purl.scheme:
        url = f"http://{url}"

    try:
        pretty.print_debug(__name__, "show_url", url)
        return ty.cast(
            "bool",
            Gtk.show_uri(
                Gdk.Screen.get_default(), url, Gtk.get_current_event_time()
            ),
        )
    except GLib.GError as exc:
        pretty.print_error(__name__, "Gtk.show_uri:", exc)

    return False


def show_help_url(url: str) -> bool:
    """Try at length to display a startup notification for the help browser.

    Return False if there is no handler for the help URL
    """
    ## Check that the system help viewer is Yelp,
    ## and if it is, launch its startup notification.
    scheme = Gio.File.new_for_uri(url).get_uri_scheme()
    default = Gio.app_info_get_default_for_uri_scheme(scheme)
    if not default:
        return False

    help_viewer_id = "yelp.desktop"

    try:
        yelp = Gio.DesktopAppInfo.new(help_viewer_id)
    except (TypeError, RuntimeError):
        return show_url(url)

    cmd_path = fileutils.lookup_exec_path(default.get_executable())
    yelp_path = fileutils.lookup_exec_path(yelp.get_executable())
    if cmd_path and yelp_path and os.path.samefile(cmd_path, yelp_path):
        with suppress(SpawnError):
            spawn_async_notify_as(help_viewer_id, [cmd_path, url])
            return True

    return show_url(url)


def get_display_path_for_bytestring(filepath: str | bytes) -> str:
    """Return a unicode path for display for string or bytestring @filepath

    Will use glib's filename decoding functions, and will
    format nicely (denote home by ~/ etc)
    """
    desc: str = GLib.filename_display_name(filepath)
    homedir = system.get_homedir()
    if desc.startswith(homedir) and homedir != desc:
        desc = f"~{desc[len(homedir) :]}"

    return desc


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations

import importlib.util
import pkgutil
import sys
import textwrap
import traceback
import types
import typing as ty
from enum import Enum

from kupfer import icons
from kupfer import plugin as kplugin
from kupfer.core import settings
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.support.types import ExecInfo

# import kupfer.icons on demand later

__all__ = (
    "LoadingError",
    "NotEnabledError",
    "PluginAttr",
    "get_plugin_attribute",
    "get_plugin_attributes",
    "get_plugin_desc",
    "get_plugin_error",
    "get_plugin_ids",
    "get_plugin_info",
    "get_plugin_name",
    "initialize_plugin",
    "is_plugin_loaded",
    "load_plugin_objects",
    "register_plugin_unimport_hook",
    "unimport_plugin",
)


class PluginAttr(Enum):
    SOURCES = "__kupfer_sources__"
    TEXT_SOURCES = "__kupfer_text_sources__"
    CONTENT_DECORATORS = "__kupfer_contents__"
    ACTION_DECORATORS = "__kupfer_actions__"
    ACTION_GENERATORS = "__kupfer_action_generators__"
    SETTINGS = "__kupfer_settings__"
    INITIALIZE = "initialize_plugin"
    FINALIZE = "finalize_plugin"

    # info attributes
    KUPFER_NAME = "__kupfer_name__"
    VERSION = "__version__"
    DESCRIPTION = "__description__"
    AUTHOR = "__author__"

    # other
    NAME = "__name__"
    FILE = "__file__"


_INFO_ATTRIBUTES = tuple(
    pa.value
    for pa in (
        PluginAttr.KUPFER_NAME,
        PluginAttr.VERSION,
        PluginAttr.DESCRIPTION,
        PluginAttr.AUTHOR,
    )
)

_PLUGIN_ICON_FILE = "icon-list"
_PLUGIN_HOOKS: dict[str, list[tuple[ty.Callable[..., None], ty.Any]]] = {}


class NotEnabledError(Exception):
    """Plugin may not be imported since it is not enabled"""


def get_plugin_ids() -> ty.Iterator[str]:
    """Enumerate possible plugin ids. Return a sequence of possible plugin ids,
    not guaranteed to be plugins."""

    for _importer, modname, _ispkg in pkgutil.iter_modules(kplugin.__path__):
        if (
            modname != "__init__"
            and not modname.endswith("_support")
            and not modname.startswith("_")
        ):
            yield modname


# pylint: disable=too-few-public-methods
class _FakePlugin:
    def __init__(
        self,
        plugin_id: str,
        attributes: dict[str, ty.Any],
        exc_info: ExecInfo | None,
    ) -> None:
        self.is_fake_plugin = True
        self.exc_info = exc_info
        self.__name__ = plugin_id
        vars(self).update(attributes)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.__name__}>"


PluginModule = ty.Union[types.ModuleType, _FakePlugin]
# imported plugins, none=not existing
_IMPORTED_PLUGINS: dict[str, PluginModule | None] = {}


def get_plugin_info() -> ty.Iterator[dict[str, ty.Any]]:
    """Generator, yields dictionaries of plugin descriptions with at least the
    fields: name, localized_name, version, description, author."""
    for plugin_name in sorted(get_plugin_ids()):
        try:
            plugin = _import_plugin_any(plugin_name)
            if not plugin:
                continue

            plugin = vars(plugin)
        except ImportError as exc:
            pretty.print_error(
                __name__, f"import plugin '{plugin_name}':", exc
            )
            continue
        except Exception:
            pretty.print_error(__name__, f"Could not load '{plugin_name}'")
            pretty.print_exc(__name__)
            continue

        localized_name = plugin.get("__kupfer_name__", None)
        desc = plugin.get("__description__", "")
        vers = plugin.get("__version__", "")
        author = plugin.get("__author__", "")
        # skip false matches;
        # all plugins have to have @localized_name
        if localized_name is None:
            continue

        yield {
            "name": plugin_name,
            "localized_name": localized_name,
            "version": vers,
            "description": desc or "",
            "author": author,
            "provides": (),
        }


def get_plugin_desc() -> str:
    """Return a formatted list of plugins suitable for printing to terminal"""
    infos = list(get_plugin_info())
    verlen = max(len(r["version"]) for r in infos)
    idlen = max(len(r["name"]) for r in infos)
    maxlen = 78
    left_margin = 2 + idlen + 1 + verlen + 1

    def format_desc(rec: dict[str, ty.Any]) -> str:
        # Wrap the description and align continued lines
        wrapped = textwrap.wrap(rec["description"], maxlen - left_margin)
        description = ("\n" + " " * left_margin).join(wrapped)
        name = rec["name"].ljust(idlen)
        ver = rec["version"].ljust(verlen)
        return f"  {name} {ver} {description}"

    return "\n".join(map(format_desc, infos))


class LoadingError(ImportError):
    pass


def _truncate_source(
    text: str, find_attributes: ty.Iterable[str]
) -> ty.Iterator[str]:
    found_info_attributes = set(find_attributes)
    for line in text.splitlines():
        # skip import from __future__ that must be in first line
        if line.startswith("from __future__ import "):
            continue

        # keep empty lines
        if not line.strip():
            yield line
            continue

        first_word = line.split(None, 1)[0]
        if first_word in found_info_attributes:
            found_info_attributes.discard(first_word)
        elif first_word in ("from", "import", "class", "def", "if"):
            break

        yield line

    if found_info_attributes:
        raise LoadingError(
            "Could not pre-load plugin: Fields missing: "
            f"{list(found_info_attributes)}. "
            "These fields need to be defined before any other code, "
            "including imports."
        )


def _import_plugin_fake(
    modpath: str, error: ExecInfo | None = None
) -> _FakePlugin | None:
    """Return an object that has the plugin info attributes we can rescue
    from a plugin raising on import.

    @error: If applicable, a tuple of exception info
    """
    try:
        spec = importlib.util.find_spec(modpath)
    except (ImportError, AttributeError, TypeError, ValueError):
        return None

    loader = spec.loader if spec is not None else None
    if not loader:
        return None

    code = loader.get_source(modpath)  # type: ignore
    if not code:
        return None

    try:
        filename = loader.get_filename(modpath)  # type: ignore
    except AttributeError:
        try:
            filename = loader.archive + loader.prefix  # type: ignore
        except AttributeError:
            filename = f"<{modpath}>"

    env = {"__name__": modpath, "__file__": filename, "__builtins__": {"_": _}}
    code = "\n".join(_truncate_source(code, _INFO_ATTRIBUTES))
    try:
        # pylint: disable=eval-used
        eval(compile(code, filename, "exec"), env)
    except Exception:
        pretty.print_error(__name__, "When loading", modpath)
        pretty.print_exc(__name__)

    attributes = {k: env.get(k) for k in _INFO_ATTRIBUTES}
    attributes.update(
        (k.value, env.get(k.value)) for k in (PluginAttr.NAME, PluginAttr.FILE)
    )
    return _FakePlugin(modpath, attributes, error)


def _import_hook_fake(pathcomps: ty.Iterable[str]) -> PluginModule | None:
    modpath = ".".join(pathcomps)
    return _import_plugin_fake(modpath)


def _import_hook_true(pathcomps: tuple[str, ...]) -> PluginModule:
    """@pathcomps path components to the import"""
    path = ".".join(pathcomps)
    fromlist = pathcomps[-1:]
    try:
        setctl = settings.get_settings_controller()
        if not setctl.get_plugin_enabled(pathcomps[-1]):
            raise NotEnabledError(f"{pathcomps[-1]} is not enabled")

        plugin = __import__(path, fromlist=fromlist)
        pretty.print_debug(__name__, f"Loading {plugin.__name__}")
        pretty.print_debug(__name__, f"  from {plugin.__file__}")
        return plugin

    except ImportError as exc:
        # Try to find a fake plugin if it exists
        fake_plugin = _import_plugin_fake(path, error=sys.exc_info())
        if not fake_plugin:
            raise

        pretty.print_error(
            __name__,
            f"Could not import plugin '{fake_plugin.__name__}': {exc}",
        )
        return fake_plugin


def _import_plugin_true(name: str) -> PluginModule | None:
    """Try to import the plugin from the package, and then from our plugin
    directories in $."""
    plugin = None
    try:
        plugin = _staged_import(name, _import_hook_true)
    except (ImportError, NotEnabledError):
        # Reraise to send this up
        raise
    except Exception:
        # catch any other error for plugins and write traceback
        traceback.print_exc()
        pretty.print_error(__name__, f"Could not import plugin '{name}'")

    return plugin


def _staged_import(
    name: str,
    import_hook: ty.Callable[[tuple[str, ...]], PluginModule | None],
) -> PluginModule | None | ty.Any:
    """Import plugin @name using @import_hook"""
    # FIXME: ty.Any because typeguard
    try:
        return import_hook(_plugin_path(name))
    except ImportError as exc:
        if name not in exc.args[0]:
            raise

    return None


def _import_plugin(name: str) -> PluginModule | None:
    if is_plugin_loaded(name):
        return _IMPORTED_PLUGINS[name]

    plugin = None
    try:
        plugin = _import_plugin_true(name)
    except NotEnabledError:
        plugin = _staged_import(name, _import_hook_fake)
    finally:
        # store non-existing plugins as None here
        _IMPORTED_PLUGINS[name] = plugin

    return plugin


def _import_plugin_any(name: str) -> ty.Any:
    if name in _IMPORTED_PLUGINS:
        return _IMPORTED_PLUGINS[name]

    return _staged_import(name, _import_hook_fake)


def _plugin_path(name: str) -> tuple[str, ...]:
    return ("kupfer", "plugin", name)


# Plugin Attributes
def get_plugin_attributes(
    plugin_name: str,
    attrs: tuple[str | PluginAttr, ...],
    warn: bool = False,
) -> ty.Iterator[ty.Any]:
    """Generator of the attributes named @attrs to be found in plugin
    @plugin_name. If the plugin is not found, we write an error and yield
    nothing. If @warn, we print a warning if a plugin does not have a requested
    attribute."""
    try:
        plugin = _import_plugin(plugin_name)
    except ImportError as exc:
        pretty.print_info(__name__, f"Skipping plugin {plugin_name}: {exc}")
        return

    for attr in attrs:
        if isinstance(attr, PluginAttr):
            attr = attr.value  # noqa: PLW2901

        try:
            yield getattr(plugin, attr)
        except AttributeError as exc:
            if warn:
                pretty.print_info(__name__, f"Plugin {plugin_name}: {exc}")

            yield None


def get_plugin_attribute(
    plugin_name: str, attr: PluginAttr | str
) -> ty.Any | None:
    """Get single plugin attribute"""
    for value in get_plugin_attributes(plugin_name, (attr,)):
        return value or None

    return None


def load_plugin_objects(
    plugin_name: str,
    attr: PluginAttr = PluginAttr.SOURCES,
    instantiate: bool = True,
) -> ty.Iterable[ty.Any]:
    """Load plugin sources or actions or other type objects (selected by @attr)."""
    objects = get_plugin_attribute(plugin_name, attr)
    if not objects:
        return

    for obj in get_plugin_attributes(plugin_name, objects, warn=True):
        if obj:
            if instantiate:
                yield obj()
            else:
                yield obj

        else:
            pretty.print_info(
                __name__, f"Object not found for {plugin_name} in {attr}"
            )


# Plugin Initialization & Error
def is_plugin_loaded(plugin_name: str) -> bool:
    if plg := _IMPORTED_PLUGINS.get(plugin_name):
        return not getattr(plg, "is_fake_plugin", None)

    return False


def _loader_hook(modpath: tuple[str, ...]) -> ty.Any:
    modname = ".".join(modpath)
    try:
        spec = importlib.util.find_spec(modname)
    except (ImportError, AttributeError, TypeError, ValueError) as ex:
        raise ImportError(f"No loader found for {modname}") from ex

    loader = spec.loader if spec is not None else None
    if not loader:
        raise ImportError(f"No loader found for {modname}")

    if not loader.is_package(modname):  # type: ignore
        raise ImportError("Is not a package")

    return loader


def _load_icons(plugin_name: str) -> None:
    try:
        _loader = _staged_import(plugin_name, _loader_hook)
    except ImportError:
        return

    modname = ".".join(_plugin_path(plugin_name))

    try:
        icon_file = pkgutil.get_data(modname, _PLUGIN_ICON_FILE)
    except OSError:
        # icon-list file just missing, let is pass silently
        return

    def get_icon_data(basename):
        return pkgutil.get_data(modname, basename)

    if icon_file:
        icons.parse_load_icon_list(icon_file, get_icon_data, plugin_name)


def initialize_plugin(plugin_name: str) -> None:
    """Initialize plugin.  Find settings attribute if defined, and initialize
    it."""
    _load_icons(plugin_name)
    if settings_dict := get_plugin_attribute(plugin_name, PluginAttr.SETTINGS):
        settings_dict.initialize(plugin_name)

    if initialize := get_plugin_attribute(plugin_name, PluginAttr.INITIALIZE):
        initialize(plugin_name)

    if finalize := get_plugin_attribute(plugin_name, PluginAttr.FINALIZE):
        register_plugin_unimport_hook(plugin_name, finalize, plugin_name)


def unimport_plugin(plugin_name: str) -> None:
    """Remove @plugin_name from the plugin list and dereference its
    python modules."""
    # Run unimport hooks
    if plugin_name in _PLUGIN_HOOKS:
        try:
            for callback, args in reversed(_PLUGIN_HOOKS[plugin_name]):
                callback(*args)

        except Exception:
            pretty.print_error(__name__, "Error finalizing", plugin_name)
            pretty.print_exc(__name__)

        _PLUGIN_HOOKS.pop(plugin_name)

    _IMPORTED_PLUGINS.pop(plugin_name)
    plugin_module_name = ".".join(_plugin_path(plugin_name))
    pretty.print_debug(__name__, "Dereferencing module", plugin_module_name)
    if plugin_module_name in sys.modules:
        sys.modules.pop(plugin_module_name)

    for mod in list(sys.modules):
        if mod.startswith(plugin_module_name + "."):
            pretty.print_debug(__name__, "Dereferencing module", mod)
            sys.modules.pop(mod)


def register_plugin_unimport_hook(
    plugin_name: str, callback: ty.Callable[..., None], *args: ty.Any
) -> None:
    if plugin_name not in _IMPORTED_PLUGINS:
        raise ValueError(f"No such plugin {plugin_name}")

    _PLUGIN_HOOKS.setdefault(plugin_name, []).append((callback, args))


def get_plugin_error(plugin_name: str) -> ty.Any:
    """Return None if plugin is loaded without error, else return a tuple of
    exception information."""
    try:
        if (plugin := _import_plugin(plugin_name)) and getattr(
            plugin, "is_fake_plugin", None
        ):
            return plugin.exc_info

    except ImportError:
        return sys.exc_info()

    return None


def get_plugin_name(modulename: str) -> str:
    """Get plugin name according to @modulename.
    (return __kupfer_name__ in @modulename module).
    """
    name: str | None = None
    for module in _IMPORTED_PLUGINS.values():
        if not module or isinstance(module, _FakePlugin):
            continue

        if module.__name__ == modulename:
            name = module.__kupfer_name__
            break

    if name:
        return name

    return modulename.split(".")[-1]



from __future__ import annotations

import typing as ty

from kupfer.core._support import get_leaf_members
from kupfer.obj import Action, AnySource, KupferObject, Leaf

if ty.TYPE_CHECKING:
    from kupfer.core.sources import SourceController


__all__ = (
    "action_valid_for_item",
    "actions_for_item",
    "iobject_source_for_action",
    "iobjects_valid_for_action",
)


def action_valid_for_item(action: Action, leaf: Leaf) -> bool:
    """Check is `action` is valid for all `leaf` representation.

    Note: valid_for_item MUST return bool.
    """
    return all(action.valid_for_item(L) for L in get_leaf_members(leaf))


def actions_for_item(
    leaf: Leaf | None, sourcecontroller: SourceController
) -> set[Action]:
    """Get list of actions for `leaf` from `sourcecontroller`."""

    if leaf is None:
        return set()

    actions: set[Action] | None = None

    for member in get_leaf_members(leaf):
        l_actions = set(member.get_actions())
        l_actions.update(sourcecontroller.get_actions_for_leaf(member))
        if actions is None:
            actions = l_actions
        else:
            actions.intersection_update(l_actions)

    return actions or set()


def iobject_source_for_action(
    action: Action, for_item: Leaf
) -> tuple[AnySource | None, bool]:
    """Get iobjects source for `action` for `for_item`.

    Simply call `Action.object_source` and 'Action.object_source_and_catalog`
    for first representation of `for_item`.

    Return (src, use_catalog)

    where
    src: object source or None,
    use_catalog: True to use catalog in addition.
    """
    leaf, *_rest = get_leaf_members(for_item)
    if not leaf:
        return None, False

    return action.object_source(leaf), action.object_source_and_catalog(leaf)


FilteringFunction = ty.Callable[
    [ty.Iterable[KupferObject]], ty.Iterable[ty.Union[Action, Leaf]]
]


def iobjects_valid_for_action(
    action: Action, for_item: Leaf
) -> FilteringFunction:
    """Return a filtering *function* that will let through
    those leaves that are good iobjects for @action and @for_item.
    """
    types = tuple(action.object_types())

    if not hasattr(action, "valid_object"):

        def type_check(items):
            return (i for i in items if isinstance(i, types))

        return type_check

    _valid_object = action.valid_object

    def type_obj_check(iobjs):
        return (
            i
            for i in iobjs
            if isinstance(i, types)
            and all(
                _valid_object(leaf, for_item=item)
                for leaf in get_leaf_members(i)
                for item in get_leaf_members(for_item)
            )
        )

    return type_obj_check



__kupfer_name__ = _("VirtualBox")
__kupfer_sources__ = ("VBoxMachinesSource",)
__description__ = _(
    "Control VirtualBox Virtual Machines. "
    "Supports both Sun VirtualBox and Open Source Edition."
)
__version__ = "0.4"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import typing as ty
from contextlib import suppress

from kupfer import plugin_support
from kupfer.obj import Action, Leaf
from kupfer.obj.apps import ApplicationSource
from kupfer.plugin.virtualbox import constants as vbox_const
from kupfer.plugin.virtualbox import ose_support
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "force_cli",
        "label": _("Force use CLI interface"),
        "type": bool,
        "value": False,
    },
)


def _get_vbox():
    if __kupfer_settings__["force_cli"]:
        pretty.print_info(__name__, "Using cli...")
        return ose_support

    with suppress(ImportError):
        # pylint: disable=import-outside-toplevel
        from kupfer.plugin.virtualbox import vboxapi4_support

        pretty.print_info(__name__, "Using vboxapi4...")
        return vboxapi4_support

    with suppress(ImportError):
        # pylint: disable=import-outside-toplevel
        from kupfer.plugin.virtualbox import vboxapi_support

        pretty.print_info(__name__, "Using vboxapi...")
        return vboxapi_support

    pretty.print_info(__name__, "Using cli...")
    return ose_support


class _VBoxSupportProxy:
    _vbox = None

    def __getattr__(self, attr):
        if not self._vbox:
            self.reload_settings()

        return getattr(self._vbox, attr)

    def reload_settings(self):
        pretty.print_debug(__name__, "_VBoxSupportProxy.reloading...")
        self.unload_module()
        self._vbox = _get_vbox()

    def unload_module(self):
        if self._vbox:
            self._vbox.unload()
            self._vbox = None

    def __bool__(self):
        return self._vbox is not None


vbox_support = _VBoxSupportProxy()


class VirtualMachine(Leaf):
    def __init__(self, obj, name, description):
        Leaf.__init__(self, obj, name)
        self.description = description

    def get_description(self):
        return self.description

    def get_icon_name(self):
        return vbox_support.ICON

    def get_actions(self):
        state = vbox_support.get_machine_state(self.object)
        if state == vbox_const.VM_STATE_POWEROFF:
            yield VMAction(
                _("Power On"), "system-run", vbox_const.VM_START_NORMAL
            )
            yield VMAction(
                _("Power On Headless"),
                "system-run",
                vbox_const.VM_START_HEADLESS,
                -5,
            )
        elif state == vbox_const.VM_STATE_POWERON:
            yield VMAction(
                _("Send Power Off Signal"),
                "system-shutdown",
                vbox_const.VM_ACPI_POWEROFF,
                -5,
            )
            yield VMAction(_("Pause"), "pause", vbox_const.VM_PAUSE)
            yield VMAction(
                _("Reboot"), "system-reboot", vbox_const.VM_REBOOT, -10
            )
        elif state == vbox_const.VM_STATE_SAVED:
            yield VMAction(
                _("Power On"), "system-run", vbox_const.VM_START_NORMAL
            )
            yield VMAction(
                _("Power On Headless"),
                "system-run",
                vbox_const.VM_START_HEADLESS,
                -5,
            )
        else:  # VM_STATE_PAUSED
            yield VMAction(_("Resume"), "resume", vbox_const.VM_RESUME)

        if state in (vbox_const.VM_STATE_POWERON, vbox_const.VM_STATE_PAUSED):
            yield VMAction(
                _("Save State"), "system-suspend", vbox_const.VM_SAVE
            )
            yield VMAction(
                _("Power Off"), "system-shutdown", vbox_const.VM_POWEROFF, -10
            )


class VMAction(Action):
    def __init__(self, name, icon, command, rank_adjust=0):
        Action.__init__(self, name)
        self._icon = icon
        self.rank_adjust = rank_adjust
        self.command = command

    def get_icon_name(self):
        return self._icon

    def item_types(self):
        yield VirtualMachine

    def activate(self, leaf, iobj=None, ctx=None):
        vbox_support.vm_action(self.command, leaf.object)


class VBoxMachinesSource(ApplicationSource):
    appleaf_content_id = ("virtualbox-ose", "virtualbox")

    def __init__(self, name=_("VirtualBox Machines")):
        super().__init__(name)
        self.monitor_token = None

    def initialize(self):
        if vbox_support.MONITORED_DIRS:
            self.monitor_token = self.monitor_directories(
                *vbox_support.MONITORED_DIRS
            )

        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )

    def finalize(self):
        if vbox_support:
            vbox_support.unload_module()

    def is_dynamic(self):
        return vbox_support.IS_DYNAMIC

    def get_items(self):
        for (
            machine_id,
            machine_name,
            machine_desc,
        ) in vbox_support.get_machines():
            yield VirtualMachine(machine_id, machine_name, machine_desc)

    def get_description(self):
        return None

    def get_icon_name(self):
        return vbox_support.ICON

    def provides(self):
        yield VirtualMachine

    def _on_setting_changed(self, _setting, _key, _value):
        if vbox_support:
            vbox_support.reload_settings()



import hashlib
import os
import pickle
import typing as ty
from pathlib import Path

from gi.repository import GdkPixbuf, Gio, GLib

from kupfer import puid
from kupfer.obj.base import KupferObject
from kupfer.support import conspickle, pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

KUPFER_COMMAND_SHEBANG = b"#!/usr/bin/env kupfer-exec\n"

__all__ = (
    "ExecutionError",
    "parse_kfcom_file",
    "save_to_file",
    "update_icon",
)


class ExecutionError(Exception):
    pass


def parse_kfcom_file(filepath: str) -> tuple[ty.Any, ...]:
    """Extract the serialized command inside @filepath

    The file must be executable (comparable to a shell script)
    >>> parse_kfcom_file(__file__)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    ExecutionError: ... (not executable)

    Return commands triple
    """
    if not os.access(filepath, os.X_OK):
        raise ExecutionError(
            _('No permission to run "%s" (not executable)')
            % GLib.filename_display_basename(filepath)
        )

    data = Path(filepath).read_bytes()

    # strip shebang away
    if data.startswith(b"#!") and b"\n" in data:
        _shebang, data = data.split(b"\n", 1)

    try:
        id_ = conspickle.BasicUnpickler.loads(data)
        command_object = puid.resolve_unique_id(id_)
    except pickle.UnpicklingError as err:
        raise ExecutionError(f"Could not parse: {err}") from err
    except Exception as err:
        raise ExecutionError(
            f'"{os.path.basename(filepath)}" is not a saved command'
        ) from err

    if command_object is None:
        raise ExecutionError(
            _('Command in "%s" is not available')
            % GLib.filename_display_basename(filepath)
        )

    try:
        return tuple(command_object.object)  # type: ignore
    except (AttributeError, TypeError) as exe:
        raise ExecutionError(
            f'"{os.path.basename(filepath)}" is not a saved command'
        ) from exe
    finally:
        GLib.idle_add(update_icon, command_object, filepath)


def save_to_file(command_leaf: ty.Any, filename: str) -> None:
    ofd = os.open(filename, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o777)
    with os.fdopen(ofd, "wb") as wfile:
        wfile.write(KUPFER_COMMAND_SHEBANG)
        pickle.dump(puid.get_unique_id(command_leaf), wfile, protocol=3)


def _write_thumbnail(gfile: Gio.File, pixbuf: GdkPixbuf.Pixbuf) -> Path:
    uri = gfile.get_uri()
    hashname = hashlib.md5(uri.encode("utf-8")).hexdigest()
    thumb_dir = Path("~/.thumbnails/normal").expanduser()
    thumb_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
    thumb_filename = thumb_dir.joinpath(hashname + ".png")
    pixbuf.savev(str(thumb_filename), "png", [], [])
    return thumb_filename


def update_icon(kobj: KupferObject, filepath: str) -> None:
    """Give @filepath a custom icon taken from @kobj"""
    icon_key = "metadata::custom-icon"

    gfile = Gio.File.new_for_path(filepath)
    finfo = gfile.query_info(icon_key, Gio.FileQueryInfoFlags.NONE, None)
    custom_icon_uri = finfo.get_attribute_string(icon_key)
    if (
        custom_icon_uri
        and Gio.File.new_for_uri(custom_icon_uri).query_exists()
    ):
        return

    namespace = gfile.query_writable_namespaces()  # FileAttributeInfoList
    if namespace.n_infos > 0:
        pretty.print_debug(__name__, "Updating icon for", filepath)
        thumb_filename = _write_thumbnail(gfile, kobj.get_pixbuf(128))
        try:
            gfile.set_attribute_string(
                icon_key,
                Gio.File.new_for_path(str(thumb_filename)).get_uri(),
                Gio.FileQueryInfoFlags.NONE,
                None,
            )
        except GLib.GError:
            pretty.print_exc(__name__)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations

import signal
import sys
import typing as ty
from contextlib import suppress

import cairo
from gi.repository import Gdk, GLib, Gtk

try:
    from gi.repository import AyatanaAppIndicator3 as AppIndicator3
except ImportError:
    AppIndicator3 = None

import kupfer.config
import kupfer.environment
from kupfer import version
from kupfer.core import commandexec, settings
from kupfer.core.datactrl import DataController
from kupfer.support import pretty, scheduler
from kupfer.ui import (
    about,
    keybindings,
    kupferhelp,
    listen,
    preferences,
    uievents,
)
from kupfer.ui._support import text_direction_is_ltr
from kupfer.ui.interface import Interface

if ty.TYPE_CHECKING:
    from gettext import gettext as _


_KUPFER_CSS: ty.Final = b"""
#kupfer {
    border: 2px solid alpha(black, 0.3);
    border-radius: 10px;
}

.matchview {
    border-radius: 0.6em;
}

#kupfer-preedit {
    padding: 0 0 0 0;
}

#kupfer-preedit.hidden {
    border-width: 0 0 0 0;
    padding: 0 0 0 0 ;
    margin: 0 0 0 0;
    outline-width: 0;
    min-height: 0;
    min-width: 0;
}

#kupfer-object-pane {
}

#kupfer-action-pane {
}

#kupfer-indirect-object-pane {
}

#kupfer-list {
}

#kupfer-list-view {
}

*:selected.matchview {
    background: alpha(@theme_selected_bg_color, 0.5);
    border: 2px solid alpha(black, 0.3)
}
"""

WINDOW_BORDER_WIDTH: ty.Final = 8


class WindowController(pretty.OutputMixin):
    """
    This is the fundamental Window (and App) Controller
    """

    def __init__(self):
        self._window: Gtk.Window = None
        self._current_screen_handler = 0
        self._interface: Interface = None  # type: ignore
        self._statusicon = None
        self._statusicon_ai = None
        self._window_hide_timer = scheduler.Timer()

    def _initialize(self, data_controller: DataController) -> None:
        self._window = Gtk.Window(
            type=Gtk.WindowType.TOPLEVEL,
            border_width=WINDOW_BORDER_WIDTH,
            decorated=False,
            name="kupfer",
        )
        # screen = self._window.get_screen()  # pylint: disable=no-member
        # if (visual := screen.get_rgba_visual()) and screen.is_composited():
        #     self._window.set_visual(visual)  # pylint: disable=no-member

        self._window.connect("realize", self._on_window_realize)
        self._window.connect(
            "configure-event", self._on_window_configure_event
        )
        self._window.add_events(  # pylint: disable=no-member
            Gdk.EventMask.BUTTON_PRESS_MASK
        )
        data_controller.connect("launched-action", self._on_launch_action)
        data_controller.connect("command-result", self._on_command_result)

        self._interface = Interface(data_controller, self._window)
        self._interface.connect("launched-action", self._on_launch_action)
        self._interface.connect("cancelled", self._on_cancelled)
        self._window.connect("map-event", self._on_window_map_event)
        self._setup_window()

        # Accept drops
        self._window.drag_dest_set(  # pylint: disable=no-member
            Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY
        )
        self._window.drag_dest_add_uri_targets()  # pylint: disable=no-member
        self._window.drag_dest_add_text_targets()  # pylint: disable=no-member
        self._window.connect("drag-data-received", self._on_drag_data_received)

        signal.signal(signal.SIGINT, self._on_sigterm)
        signal.signal(signal.SIGTERM, self._on_sigterm)
        signal.signal(signal.SIGHUP, self._on_sigterm)

    def _on_window_map_event(self, *_args: ty.Any) -> None:
        self._interface.update_third()

    def _on_window_realize(self, widget: Gtk.Widget) -> None:
        # Load css
        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(_KUPFER_CSS)

        Gtk.StyleContext.add_provider_for_screen(  # pylint: disable=no-member
            widget.get_screen(),
            style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

    def _on_window_configure_event(
        self, widget: Gtk.Widget, event: Gdk.EventConfigure
    ) -> None:
        """Set main window rounded shape."""
        width, height = widget.get_size()
        radius = 10
        width_radius = width - radius
        height_radius = height - radius
        surface = cairo.ImageSurface(  # pylint: disable=no-member
            cairo.FORMAT_ARGB32,  # pylint: disable=no-member
            width,
            height,
        )
        ctx = cairo.Context(surface)  # pylint: disable=no-member
        ctx.set_source_rgba(0.0, 0.0, 255.0, 128.0)
        ctx.set_operator(cairo.OPERATOR_SOURCE)  # pylint: disable=no-member
        ctx.move_to(radius, 0)
        ctx.line_to(width_radius, 0)
        ctx.arc(width_radius, radius, radius, 4.71, 6.28)
        ctx.line_to(width, height_radius)
        ctx.arc(width_radius, height_radius, radius, 0, 1.57)
        ctx.line_to(radius, height)
        ctx.arc(radius, height_radius, radius, 1.57, 3.14)
        ctx.line_to(0, radius)
        ctx.arc(radius, radius, radius, 3.14, 4.71)
        ctx.close_path()
        ctx.fill()
        reg = Gdk.cairo_region_create_from_surface(surface)
        widget.shape_combine_region(reg)

    def _show_statusicon(self) -> None:
        """Create (if not exists) and show status icon."""
        if not self._statusicon:
            status = Gtk.StatusIcon.new_from_icon_name(version.ICON_NAME)
            status.set_tooltip_text(version.PROGRAM_NAME)
            status.connect(
                "popup-menu", self._on_popup_menu, self._setup_menu()
            )
            status.connect("activate", self._on_statusicon_activate)
            self._statusicon = status

        with suppress(AttributeError):
            self._statusicon.set_visible(True)

    def _on_showstatusicon_changed(
        self,
        setctl: settings.SettingsController,
        section: str,
        key: str,
        value: ty.Any,
    ) -> None:
        """Callback from SettingsController - show / hide status icon on
        settings changed."""
        if value:
            self._show_statusicon()
        elif self._statusicon:
            # hide
            try:
                self._statusicon.set_visible(False)
            except AttributeError:
                self._statusicon = None

    def _show_statusicon_ai(self) -> None:
        """Show (create if not exists) status icon using AyatanaAppIndicator3."""
        if AppIndicator3 is None:
            return

        if not self._statusicon_ai:
            indicator = AppIndicator3.Indicator.new(
                version.PROGRAM_NAME,
                version.ICON_NAME,
                AppIndicator3.IndicatorCategory.APPLICATION_STATUS,
            )
            indicator.set_menu(self._setup_menu())
            self._statusicon_ai = indicator

        self._statusicon_ai.set_status(AppIndicator3.IndicatorStatus.ACTIVE)

    def _showstatusicon_ai_changed(
        self,
        setctl: settings.SettingsController,
        section: str,
        key: str,
        value: ty.Any,
    ) -> None:
        """Show/Hide AppIndicator3 status icon on preferences change."""
        if AppIndicator3 is None:
            return

        if value:
            self._show_statusicon_ai()
        elif self._statusicon_ai:
            # hide
            self._statusicon_ai.set_status(
                AppIndicator3.IndicatorStatus.PASSIVE
            )

    def _setup_menu(self, context_menu: bool = False) -> Gtk.Menu:
        menu = Gtk.Menu()
        menu.set_name("kupfer-menu")

        def submenu_callback(
            menuitem: Gtk.MenuItem, callback: ty.Callable[[], None]
        ) -> bool:
            callback()
            return True

        def add_menu_item(
            icon: str | None,
            callback: ty.Callable[..., None],
            label: str | None = None,
            with_ctx: bool = True,
        ) -> None:
            def mitem_handler(
                menuitem: Gtk.MenuItem, callback: ty.Callable[..., None]
            ) -> bool:
                if with_ctx:
                    ui_ctx = uievents.gui_context_from_widget(
                        Gtk.get_current_event_time(), menuitem
                    )
                    callback(ui_ctx)
                else:
                    callback()

                if context_menu:
                    self.put_away()

                return True

            if label and not icon:
                mitem = Gtk.MenuItem(label=label)
            else:
                mitem = Gtk.ImageMenuItem.new_from_stock(icon)

            mitem.connect("activate", mitem_handler, callback)
            menu.append(mitem)

        if context_menu:
            add_menu_item(Gtk.STOCK_CLOSE, self.put_away, with_ctx=False)
        else:
            add_menu_item(None, self._on_activate, _("Show Main Interface"))

        menu.append(Gtk.SeparatorMenuItem())
        if context_menu:
            for name, func in self._interface.get_context_actions():
                mitem = Gtk.MenuItem(label=name)
                mitem.connect("activate", submenu_callback, func)
                menu.append(mitem)

            menu.append(Gtk.SeparatorMenuItem())

        add_menu_item(Gtk.STOCK_PREFERENCES, preferences.show_preferences)
        add_menu_item(Gtk.STOCK_HELP, kupferhelp.show_help)
        add_menu_item(Gtk.STOCK_ABOUT, about.show_about_dialog)
        menu.append(Gtk.SeparatorMenuItem())
        add_menu_item(Gtk.STOCK_QUIT, self._on_quit, with_ctx=False)
        menu.show_all()

        return menu

    def _setup_window(self) -> None:
        """Returns window."""
        self._window.connect("delete-event", self._on_close_window)
        self._window.connect("focus-out-event", self._on_lost_focus)
        self._window.connect(
            "button-press-event", self._on_window_frame_clicked
        )
        widget = self._interface.get_widget()
        widget.show()

        # Build the window frame with its top bar
        topbar = Gtk.HBox()

        vbox = Gtk.VBox()
        vbox.pack_start(topbar, False, False, 0)
        vbox.pack_start(widget, True, True, 0)
        vbox.show()
        self._window.add(vbox)  # pylint: disable=no-member

        title = Gtk.Label.new("")
        button = Gtk.Label.new("")
        l_programname = version.PROGRAM_NAME.lower()
        # The text on the general+context menu button
        btext = f"<b>{l_programname} ⚙</b>"
        button.set_markup(btext)
        button_box = Gtk.EventBox()
        button_box.set_visible_window(False)
        button_adj = Gtk.Alignment.new(0.5, 0.5, 0, 0)
        button_adj.set_padding(0, 2, 0, 3)
        button_adj.add(button)
        button_box.add(button_adj)
        button_box.connect("button-press-event", self._on_context_menu_clicked)
        button_box.connect(
            "enter-notify-event", self._on_button_enter, button, btext
        )
        button_box.connect(
            "leave-notify-event", self._on_button_leave, button, btext
        )
        button.set_name("kupfer-menu-button")
        title_align = Gtk.Alignment.new(0, 0.5, 0, 0)
        title_align.add(title)
        topbar.pack_start(title_align, True, True, 0)
        topbar.pack_start(button_box, False, False, 0)
        topbar.show_all()

        self._window.set_title(version.PROGRAM_NAME)
        self._window.set_icon_name(version.ICON_NAME)
        self._window.set_type_hint(  # pylint: disable=no-member
            self._create_window_type_hint()
        )
        self._window.set_property("skip-taskbar-hint", True)
        self._window.set_keep_above(True)  # pylint: disable=no-member
        if (pos := self._create_window_position()) != Gtk.WindowPosition.NONE:
            self._window.set_position(pos)  # pylint: disable=no-member

        if not text_direction_is_ltr():
            self._window.set_gravity(  # pylint: disable=no-member
                Gdk.GRAVITY_NORTH_EAST
            )
        # Setting not resizable changes from utility window
        # on metacity
        self._window.set_resizable(False)

    def _create_window_type_hint(self) -> Gdk.WindowTypeHint:
        """Create window type hints according to WINDOW_TYPE_HINT env setting."""
        type_hint = Gdk.WindowTypeHint.UTILITY
        hint_name = kupfer.config.get_kupfer_env("WINDOW_TYPE_HINT").upper()
        if not hint_name:
            return type_hint

        if (
            hint_enum := getattr(Gdk.WindowTypeHint, hint_name, None)
        ) is not None:
            return hint_enum

        self.output_error("No such Window Type Hint", hint_name)
        self.output_error("Existing type hints:")
        for name in dir(Gdk.WindowTypeHint):
            if name.upper() == name:
                self.output_error(name)

        return type_hint

    def _create_window_position(self) -> Gtk.WindowPosition:
        """Create window position according to WINDOW_POSITION env setting."""
        value = Gtk.WindowPosition.NONE
        hint_name = kupfer.config.get_kupfer_env("WINDOW_POSITION").upper()
        if not hint_name:
            return value

        if hint_enum := getattr(Gtk.WindowPosition, hint_name, None):
            return hint_enum

        self.output_error("No such Window Position", hint_name)
        self.output_error("Existing values:")
        for name in dir(Gtk.WindowPosition):
            if name.upper() == name:
                self.output_error(name)

        return value

    def _on_window_frame_clicked(
        self, widget: Gtk.Widget, event: Gdk.EventButton
    ) -> None:
        """Start drag when the window is clicked"""
        widget.begin_move_drag(
            event.button, int(event.x_root), int(event.y_root), event.time
        )

    def _on_context_menu_clicked(
        self, widget: Gtk.Widget, event: Gdk.EventButton
    ) -> bool:
        """The context menu label was clicked"""
        menu = self._setup_menu(True)
        menu.set_screen(self._window.get_screen())  # pylint: disable=no-member
        menu.popup(None, None, None, None, event.button, event.time)
        return True

    def _on_button_enter(
        self,
        widget: Gtk.Widget,
        event: Gdk.EventCrossing,
        button: Gtk.Widget,
        udata: str,
    ) -> None:
        """Pointer enters context menu button"""
        button.set_markup(f"<u>{udata}</u>")

    def _on_button_leave(
        self,
        widget: Gtk.Widget,
        event: Gdk.EventCrossing,
        button: Gtk.Widget,
        udata: str,
    ) -> None:
        """Pointer leaves context menu button"""
        button.set_markup(udata)

    def _on_popup_menu(
        self,
        status_icon: Gtk.StatusIcon,
        button: int,
        activate_time: float,
        menu: Gtk.Menu,
    ) -> None:
        """When the StatusIcon is right-clicked."""
        menu.popup(
            None,
            None,
            Gtk.StatusIcon.position_menu,
            status_icon,
            button,
            activate_time,
        )

    def _on_launch_action(self, sender: ty.Any, *_args: ty.Any) -> None:
        """Callback for launch-action event: hide window.

        Separate window hide from the action being done. This is to solve
        a window focus bug when we switch windows using an action.
        """
        self._interface.did_launch()
        self._window_hide_timer.set_ms(100, self.put_away)

    def _on_command_result(
        self,
        sender: DataController,
        result_type: int,
        ui_ctx: uievents.GUIEnvironmentContext,
    ) -> None:
        """Callback for command-result event: update window."""
        result_type = commandexec.ExecResult(result_type)
        # handle "refresh" result
        if result_type == commandexec.ExecResult.REFRESH:
            self.output_debug("refresh leaves list")
            self._interface.reset_current()
            self._interface.did_launch()
            return

        self._interface.did_get_result()
        if ui_ctx:
            self._on_present(
                sender, ui_ctx.get_display(), ui_ctx.get_timestamp()
            )
        else:
            self._on_present(sender, "", Gtk.get_current_event_time())

    def _on_lost_focus(
        self, window: Gtk.Window, event: Gdk.EventFocus
    ) -> None:
        """Hide window on lost focus."""
        if not kupfer.config.has_capability("HIDE_ON_FOCUS_OUT"):
            return
        # Close at unfocus.
        # Since focus-out-event is triggered even when we click inside the
        # window, we'll do some additional math to make sure that
        # that window won't close if the mouse pointer is over it.
        _gdkwindow, x, y, _mods = (
            window.get_screen().get_root_window().get_pointer()
        )
        w_x, w_y = window.get_position()
        w_w, w_h = window.get_size()
        if not (w_x <= x <= w_x + w_w and w_y <= y <= w_y + w_h):
            self._window_hide_timer.set_ms(50, self.put_away)

    def _on_monitors_changed(self, *_ignored: ty.Any) -> None:
        """Center window on monitor changed."""
        self._center_window()

    def _window_put_on_screen(self, screen: Gdk.Screen) -> None:
        """Move window to `screen`."""
        if self._current_screen_handler:
            scr = self._window.get_screen()  # pylint: disable=no-member
            scr.disconnect(self._current_screen_handler)

        self._window.set_screen(screen)  # pylint: disable=no-member
        self._current_screen_handler = screen.connect(
            "monitors-changed", self._on_monitors_changed
        )

    def _center_window(self, displayname: str | None = None) -> None:
        """Center Window on the monitor the pointer is currently on"""
        # pylint: disable=no-member
        if not displayname and self._window.has_screen():
            display = self._window.get_display()
        else:
            display = uievents.GUIEnvironmentContext.ensure_display_open(
                displayname
            )

        screen, x, y, _mod = display.get_pointer()
        self._window_put_on_screen(screen)
        monitor_nr = screen.get_monitor_at_point(x, y)
        geo = screen.get_monitor_geometry(monitor_nr)
        wid, hei = self._window.get_size()
        midx = geo.x + geo.width / 2 - wid / 2
        midy = geo.y + geo.height / 2 - hei / 2

        self._window.move(midx, midy)
        uievents.try_close_unused_displays(screen)

    def _should_recenter_window(self) -> bool:
        """Return True if the mouse pointer and the window are on different
        monitors."""
        # Check if the GtkWindow was realized yet
        if not self._window.get_realized():
            return True

        # pylint: disable=no-member
        display = self._window.get_screen().get_display()
        screen, x, y, _modifiers = display.get_pointer()
        mon_cur = screen.get_monitor_at_point(x, y)
        mon_win = screen.get_monitor_at_window(self._window.get_window())
        return mon_cur != mon_win  # type: ignore

    def _on_activate(self, sender: ty.Any = None) -> None:
        """Activate (show) main windows."""
        # pylint: disable=no-member
        dispname = self._window.get_screen().make_display_name()
        self._on_present(sender, dispname, Gtk.get_current_event_time())

    def _on_present(
        self, sender: ty.Any, display: str | None, timestamp: int
    ) -> None:
        """Present on @display, where None means default display"""
        self._window_hide_timer.invalidate()
        display = display or Gdk.Display.get_default().get_name()
        # Center window before first show
        if not self._window.get_realized():
            self._center_window(display)

        self._window.stick()  # pylint: disable=no-member
        self._window.present_with_time(timestamp)
        # pylint: disable=no-member
        self._window.get_window().focus(timestamp=timestamp)
        self._interface.focus()

        # Center after present if we are moving between monitors
        if self._should_recenter_window():
            self._center_window(display)

    def put_away(self) -> None:
        """Hide main window."""
        self._interface.put_away()
        self._window.hide()

    def _on_cancelled(self, _obj: Interface) -> None:
        self.put_away()

    def _on_show_hide(
        self, sender: ty.Any, display: str, timestamp: int
    ) -> None:
        """Toggle activate/put-away."""
        if self._window.get_property("visible"):
            self.put_away()
        else:
            self._on_present(sender, display, timestamp)
            # refresh current match (if has qf_id)
            self._interface.re_search_kpfer()

    def _on_statusicon_activate(self, sender: Gtk.StatusIcon) -> None:
        """GtkStatusIcon callback"""
        self._on_show_hide(sender, "", Gtk.get_current_event_time())

    def _on_key_binding(
        self,
        keyobj: keybindings.KeyboundObject,
        keybinding_number: int,
        display: str,
        timestamp: int,
    ) -> None:
        """Keybinding activation callback"""
        if keybinding_number == keybindings.KEYBINDING_TARGET_DEFAULT:
            self._on_show_hide(keyobj, display, timestamp)

        elif keybinding_number == keybindings.KEYBINDING_TARGET_MAGIC:
            self._on_present(keyobj, display, timestamp)
            self._interface.select_selected_text()
            self._interface.select_selected_file()

    def _on_drag_data_received(
        self,
        widget: Gtk.Widget,
        context: ty.Any,
        x: int,
        y: int,
        data_: Gtk.SelectionData,
        info: ty.Any,
        time: int,
    ) -> None:
        if uris := data_.get_uris():
            self._interface.put_files(uris, paths=False)
        else:
            self._interface.put_text(data_.get_text())

    def _on_put_text(
        self, sender: Gtk.Widget, text: str, display: str, timestamp: int
    ) -> None:
        """We got a search text from dbus"""
        self._on_present(sender, display, timestamp)
        self._interface.put_text(text)

    def _on_put_files(
        self,
        sender: ty.Any,
        fileuris: ty.Iterable[str],
        display: str,
        timestamp: int,
    ) -> None:
        self._on_present(sender, display, timestamp)
        self._interface.put_files(fileuris, paths=True)

    def _on_find_object(
        self, sender: ty.Any, qpfer: str, display: str, timestamp: int
    ) -> None:
        self._on_present(sender, display, timestamp)
        self._interface.find_object(qpfer)

    def _on_execute_file(
        self,
        sender: ty.Any,
        filepath: str,
        display: str,
        timestamp: int,
    ) -> None:
        self._interface.execute_file(filepath, display, timestamp)

    def _on_close_window(self, window: Gtk.Widget, event: ty.Any) -> bool:
        """Callback for main window delete-event."""
        self.put_away()
        return True

    def _on_sigterm(self, signal_: int, _frame: ty.Any) -> None:
        self.output_info("Caught signal", signal_, "exiting..")
        self._on_quit()

    def _on_early_interrupt(self, _signal: int, _frame: ty.Any) -> None:
        sys.exit(1)

    def save_data(self) -> None:
        """Save state before quit"""
        sch = scheduler.get_scheduler()
        sch.finish()
        self._interface.save_config()

    def _on_quit(self, sender: Gtk.Widget | None = None) -> None:
        Gtk.main_quit()

    def _quit_now(self) -> None:
        """Quit immediately (state save should already be done)"""
        raise SystemExit

    def _on_session_save(self, *_args: ty.Any) -> bool:
        """Old-style session save callback.  Return True on successful."""
        # No quit, only save
        self.output_info("Saving for logout...")
        self.save_data()
        return True

    def _on_session_die(self, *_args: ty.Any) -> None:
        """Session callback on session end: quit now, without saving, since we
        already do that on Session save!
        """
        self._quit_now()

    def _lazy_setup(self, activate: bool = False) -> None:
        """Do all setup that can be done after showing main interface.
        Connect to desktop services (keybinding callback, session logout
        callbacks etc).

        When `activate`, call _on_activate at the end (show main window).
        """
        # pylint: disable=import-outside-toplevel
        from kupfer.ui import session

        self.output_debug("in lazy_setup")

        setctl = settings.get_settings_controller()
        if setctl.get_show_status_icon():
            self._show_statusicon()

        if setctl.get_show_status_icon_ai():
            self._show_statusicon_ai()

        setctl.connect(
            "value-changed::kupfer.showstatusicon",
            self._on_showstatusicon_changed,
        )
        setctl.connect(
            "value-changed::kupfer.showstatusicon_ai",
            self._showstatusicon_ai_changed,
        )

        if keystr := setctl.get_keybinding():
            succ = keybindings.bind_key(keystr)
            self.output_info(
                f"Trying to register {keystr} to spawn kupfer.. "
                + ("success" if succ else "failed")
            )

        if magickeystr := setctl.get_magic_keybinding():
            succ = keybindings.bind_key(
                magickeystr, keybindings.KEYBINDING_TARGET_MAGIC
            )
            self.output_debug(
                f"Trying to register {magickeystr} to spawn kupfer.. "
                + ("success" if succ else "failed")
            )

        keyobj = keybindings.get_keybound_object()
        keyobj.connect("keybinding", self._on_key_binding)

        client = session.SessionClient()
        client.connect("save-yourself", self._on_session_save)
        client.connect("die", self._on_session_die)
        self._interface.lazy_setup()

        if activate:
            self._on_activate()

        self.output_debug("finished lazy_setup")

    def main(self, quiet: bool = False) -> None:
        """Start WindowController, present its window (if not @quiet)"""
        signal.signal(signal.SIGINT, self._on_early_interrupt)

        kserv1 = None
        kserv2 = None

        try:
            # NOTE: For a *very short* time we will use both APIs  TODO: off one
            kserv1 = listen.Service()
            kserv2 = listen.ServiceNew()
        except listen.AlreadyRunningError:
            self.output_info("An instance is already running, exiting...")
            self._quit_now()
        except listen.NoConnectionError:
            pass

        if kserv1:
            keyobj = keybindings.get_keybound_object()
            keyobj.connect(
                "bound-key-changed",
                lambda _x, y, z: kserv1.BoundKeyChanged(y, z),
            )
            kserv1.connect("relay-keys", keyobj.relayed_keys)

        # Load data
        data_controller = DataController.instance()
        sch = scheduler.get_scheduler()
        sch.load()
        # Now create UI and display
        self._initialize(data_controller)
        sch.display()

        for kserv in (kserv1, kserv2):
            if kserv:
                kserv.connect("present", self._on_present)
                kserv.connect("show-hide", self._on_show_hide)
                kserv.connect("put-text", self._on_put_text)
                kserv.connect("put-files", self._on_put_files)
                kserv.connect("find-object", self._on_find_object)
                kserv.connect("execute-file", self._on_execute_file)
                kserv.connect("quit", self._on_quit)

        GLib.idle_add(self._lazy_setup, not quiet)

        def do_main_iterations(max_events=0):
            # use sentinel form of iter
            for idx, _pending in enumerate(iter(Gtk.events_pending, False)):
                if max_events and idx > max_events:
                    break

                Gtk.main_iteration()

        try:
            Gtk.main()
            # put away window *before exiting further*
            self.put_away()
            do_main_iterations(10)
        finally:
            self.save_data()

        # tear down but keep hanging
        for kserv in (kserv1, kserv2):
            if kserv:
                kserv.unregister()

        keybindings.bind_key(None, keybindings.KEYBINDING_TARGET_DEFAULT)
        keybindings.bind_key(None, keybindings.KEYBINDING_TARGET_MAGIC)

        do_main_iterations(100)
        # if we are still waiting, print a message
        if Gtk.events_pending():
            self.output_info("Waiting for tasks to finish...")
            do_main_iterations()



import operator
import typing as ty
from collections import defaultdict

Song = dict[str, ty.Any]


def _sort_album(album: list[Song]) -> None:
    """Sort album in track order"""

    def get_track_number(rec):
        try:
            tnr = int(rec["track-number"])
        except (KeyError, ValueError):
            tnr = 0
        return tnr

    album.sort(key=get_track_number)


def _sort_album_order(songs: list[Song]) -> None:
    """Sort songs in order by album then by track number

    >>> songs = [
    ... {"title": "a", "album": "B", "track-number": 2},
    ... {"title": "b", "album": "A", "track-number": 1},
    ... {"title": "c", "album": "B", "track-number": 1},
    ... ]
    >>> _sort_album_order(songs)
    >>> [s["title"] for s in songs]
    ['b', 'c', 'a']
    """

    songs.sort(key=operator.itemgetter("date", "album", "track-number"))


def parse_rhythmbox_albums(songs: ty.Iterable[Song]) -> dict[str, list[Song]]:
    albums: dict[str, list[Song]] = defaultdict(list)
    for song in songs:
        if not song["artist"]:
            continue

        if song_album := song["album"]:
            albums[song_album].append(song)

    # sort album in track order
    for album in albums.values():
        _sort_album(album)

    return albums


def parse_rhythmbox_artists(songs: ty.Iterable[Song]) -> dict[str, list[Song]]:
    artists: dict[str, list[Song]] = defaultdict(list)
    for song in songs:
        if song_artist := song["artist"]:
            artists[song_artist].append(song)

    # sort in album + track order
    for artist in artists.values():
        _sort_album_order(artist)

    return artists


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import typing as ty

if ty.TYPE_CHECKING:
    from gettext import gettext as _


ACCELERATOR_NAMES: ty.Final = {
    # TRANS: Names of accelerators in the interface
    "activate": _("Alternate Activate"),
    # TRANS: The "Comma Trick"/"Put Selection on Stack" allows the
    # TRANS: user to select many objects to be used for one action
    "comma_trick": _("Comma Trick"),
    # TRANS: "Compose Command" makes one object out of the selected
    # TRANS: object + action (+iobject)
    "compose_action": _("Compose Command"),
    "mark_as_default": _("Mark Default Action"),
    "erase_affinity_for_first_pane": _("Forget Object"),
    "reset_all": _("Reset All"),
    "select_quit": _("Select Quit"),
    "select_selected_file": _("Select Selected File"),
    "select_selected_text": _("Select Selected Text"),
    "select_clipboard_file": _("Select Clipboard File"),
    "select_clipboard_text": _("Select Clipboard Text"),
    "show_help": _("Show Help"),
    "show_preferences": _("Show Preferences"),
    "switch_to_source": _("Switch to 1st Pane"),
    "switch_to_2": _("Switch to 2nd Pane"),
    "switch_to_3": _("Switch to 3rd Pane"),
    "toggle_text_mode_quick": _("Toggle Text Mode"),
}



"""
User Interface Utility Functions for Kupfer

These helper functions can be called from plugins (are meant to serve this
purpose), but care should be taken to only call UI functions from the main
(default) thread.

"""

from __future__ import annotations

import math
import textwrap
import typing as ty

from gi.repository import Gdk, Gtk, Pango

from kupfer import config, version
from kupfer.support import pretty
from kupfer.ui import uievents

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec


def _window_close_on_escape(widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
    """Callback function for Window's key press event, will destroy window
    on escape"""
    if event.keyval == Gdk.keyval_from_name("Escape"):
        widget.close()
        return True

    return False


def builder_get_objects_from_file(
    fname: str, attrs: ty.Iterable[str], autoconnect_to: ty.Any = None
) -> ty.Iterator[Gtk.Widget]:
    """Open @fname with Gtk.Builder and yield objects named @attrs

    @fname is sought in the data directories.
    If @autoconnect_to is not None, signals are autoconnected to this object,
    and a user_data object is passed as a namespace containing all @attrs
    """
    builder = Gtk.Builder()
    builder.set_translation_domain(version.PACKAGE_NAME)

    ui_file = config.get_data_file(fname)
    builder.add_from_file(ui_file)

    # pylint: disable=too-few-public-methods
    class Namespace:
        pass

    names = Namespace()
    for attr in attrs:
        obj = builder.get_object(attr)
        setattr(names, attr, obj)
        yield obj

    if autoconnect_to:
        assert hasattr(autoconnect_to, "names")
        autoconnect_to.names = names
        builder.connect_signals(autoconnect_to)  # pylint: disable=no-member


class _ResultWindowBehavior:
    def __init__(self):
        # this is dynamic builder_get_objects_from_file.Namespace
        self.names: ty.Any = None

    def on_text_result_window_key_press_event(
        self, widget: Gtk.Widget, event: Gdk.EventKey
    ) -> bool:
        return _window_close_on_escape(widget, event)

    def on_close_button_clicked(self, widget: Gtk.Widget) -> bool:
        self.names.text_result_window.close()
        return True

    def on_copy_button_clicked(self, widget: Gtk.Widget) -> None:
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        textview = self.names.result_textview
        buf = textview.get_buffer()
        buf.select_range(*buf.get_bounds())
        buf.copy_clipboard(clip)


# pylint: disable=too-many-locals
def _calculate_window_size(
    window: Gtk.Window, textview: Gtk.TextView
) -> tuple[int, int]:
    # Find the size of one (monospace) letter
    playout = textview.create_pango_layout("X")
    ink_r, _logical_r = playout.get_pixel_extents()

    # Fix Sizing:
    # We want to size the window so that the
    # TextView is displayed without scrollbars
    # initially, if it fits on screen.
    tw_sr = textview.get_size_request()
    oldwid, oldhei = tw_sr.width, tw_sr.height
    winwid, winhei = window.get_size()

    # max_hsize, max_vsize = window.get_default_size()
    tw_sr = textview.size_request()
    wid, hei = tw_sr.width, tw_sr.height

    # Set max window size to 100 columns x 60 lines
    max_hsize = ink_r.height * 60
    max_vsize = ink_r.width * 100

    vsize = int(min(hei + (winhei - oldhei) + 5, max_vsize))
    hsize = int(min(wid + (winwid - oldwid) + 5, max_hsize))
    return vsize, hsize


def show_text_result(
    text: str,
    title: str | None = None,
    ctx: commandexec.ExecutionToken | None = None,
) -> None:
    """Show @text in a result window.

    Use @title to set a window title
    """

    window, textview = tuple(
        builder_get_objects_from_file(
            "result.ui",
            ("text_result_window", "result_textview"),
            autoconnect_to=_ResultWindowBehavior(),
        )
    )

    # Set up text buffer
    buf = Gtk.TextBuffer()
    buf.set_text(text)
    textview.set_buffer(buf)
    textview.set_wrap_mode(Gtk.WrapMode.NONE)
    textview.set_editable(True)

    if title:
        window.set_title(title)

    if ctx:
        ctx.environment.present_window(window)

    window.show_all()

    hsize, vsize = _calculate_window_size(window, textview)
    textview.set_wrap_mode(Gtk.WrapMode.WORD)
    window.resize(hsize, vsize)
    if ctx:
        ctx.environment.present_window(window)
    else:
        window.present_with_time(uievents.current_event_time())


def _wrap_paragraphs(text: str) -> str:
    """Return @text with linewrapped paragraphs"""
    return "\n\n".join(textwrap.fill(par) for par in text.split("\n\n"))


def _set_font_size(label: Gtk.Label, fontsize: float = 48.0) -> None:
    label.modify_font(Pango.FontDescription.from_string(str(fontsize)))


_MAX_LINE_WIDTH: ty.Final[int] = 100


def show_large_type(
    text: str, ctx: commandexec.ExecutionToken | None = None
) -> None:
    """Show @text, large, in a result window."""
    text = text.strip()

    window = Gtk.Window()
    label = Gtk.Label()
    label.set_text(text)
    label.show()

    size = 72.0
    _set_font_size(label, size)

    if ctx:
        screen = ctx.environment.get_screen()
        window.set_screen(screen)  # pylint: disable=no-member
    else:
        screen = Gdk.Screen.get_default()

    assert screen
    maxwid = screen.get_width() - 50
    maxhei = screen.get_height() - 100
    l_sr = label.size_request()  # pylint: disable=no-member
    wid, hei = l_sr.width, l_sr.height

    # If the text contains long lines, we try to
    # hard-wrap the text
    if (wid > maxwid or hei > maxhei) and any(
        len(L) > _MAX_LINE_WIDTH for L in text.splitlines()
    ):
        label.set_text(_wrap_paragraphs(text))

    l_sr = label.size_request()  # pylint: disable=no-member
    wid, hei = l_sr.width, l_sr.height
    if wid > maxwid or hei > maxhei:
        # Round size down to fit inside
        wscale = maxwid * 1.0 / wid
        hscale = maxhei * 1.0 / hei
        _set_font_size(label, math.floor(min(wscale, hscale) * size) or 1.0)

    window.add(label)  # pylint: disable=no-member
    window.set_position(Gtk.WindowPosition.CENTER)  # pylint: disable=no-member
    window.set_resizable(False)  # pylint: disable=no-member
    window.set_decorated(False)
    window.set_property("border-width", 10)
    # pylint: disable=no-member
    window.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse("black"))
    # pylint: disable=no-member
    label.modify_fg(Gtk.StateType.NORMAL, Gdk.color_parse("white"))

    def _on_key_press(widget, event):
        widget.destroy()
        return True

    window.connect("key-press-event", _on_key_press)
    window.show_all()  # pylint: disable=no-member
    if ctx:
        ctx.environment.present_window(window)
    else:
        window.present_with_time(uievents.current_event_time())


_SERVICE_NAME = "org.freedesktop.Notifications"
_OBJECT_PATH = "/org/freedesktop/Notifications"
_IFACE_NAME = "org.freedesktop.Notifications"


def _get_notification_obj() -> ty.Any:
    """we will activate it over d-bus (start if not running)"""
    # pylint: disable=import-outside-toplevel
    import dbus

    try:
        bus = dbus.SessionBus()
        proxy_obj = bus.get_object(_SERVICE_NAME, _OBJECT_PATH)
    except dbus.DBusException as exc:
        pretty.print_debug(__name__, exc)
        return None

    return proxy_obj


def show_notification(
    title: str, text: str = "", icon_name: str = "", nid: int = 0
) -> int | None:
    """Show notification with `title` and optional `text` and `icon_name`.
    `nid` is id of then notification to replace; set 0 to get new notification
    id.
    Returns the id of the displayed notification.
    """
    if not (notifications := _get_notification_obj()):
        return None

    hints = {
        "desktop-entry": version.DESKTOP_ID,
    }
    rid = notifications.Notify(
        "kupfer",
        nid,
        icon_name,
        title,
        text,
        (),
        hints,
        -1,
        dbus_interface=_IFACE_NAME,
    )
    return int(rid)


def confirm_dialog(msg: str, positive_btn: str) -> bool:
    """Request user confirmation."""
    dlg = Gtk.MessageDialog(
        None, Gtk.DialogFlags.MODAL, Gtk.MessageType.QUESTION
    )
    dlg.set_markup(msg)
    dlg.set_title(_("Kupfer"))
    dlg.add_buttons(
        Gtk.STOCK_CANCEL,
        Gtk.ResponseType.CLOSE,
        positive_btn,
        Gtk.ResponseType.ACCEPT,
    )
    dlg.set_keep_above(True)
    # pylint: disable=no-member
    result: bool = dlg.run() == Gtk.ResponseType.ACCEPT
    dlg.destroy()
    return result



"""
Support object/functions that are independent from other parts of Kupfer.
"""

from __future__ import annotations

from kupfer.support import desktop_parse

__all__ = ("argv_for_commandline",)


def argv_for_commandline(cli: str) -> list[str]:
    return desktop_parse.parse_argv(cli)



__kupfer_name__ = _("Trash")
__kupfer_actions__ = ("MoveToTrash", "EmptyTrash")
__kupfer_sources__ = ("TrashSource",)
__description__ = _("Access trash contents")
__version__ = "2017.2"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import typing as ty
from contextlib import suppress

from gi.repository import Gio, GLib

from kupfer import icons, launch
from kupfer.obj import (
    Action,
    FileLeaf,
    Leaf,
    OperationError,
    Source,
    SourceLeaf,
)
from kupfer.obj.fileactions import Open
from kupfer.support import pretty
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext


_TRASH_URI = "trash://"


class MoveToTrash(Action):
    # this should never be default
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, _("Move to Trash"))

    def activate(self, leaf, iobj=None, ctx=None):
        if not uiutils.confirm_dialog(
            _("Move file %s to trash?") % leaf.object, _("Move")
        ):
            return

        gfile = leaf.get_gfile()
        try:
            gfile.trash()
        except GLib.Error as exc:
            # pylint: disable=no-member
            raise OperationError(exc.message) from exc

    def activate_multiple(self, objs, iobjects=None, ctx=None):
        objs = list(objs)
        objs_count = len(objs)
        if not uiutils.confirm_dialog(
            ngettext(
                "Move one file to trash?",
                "Move %(num)s files to trash?",
                objs_count,
            )
            % {"num": objs_count},
            _("Move"),
        ):
            return

        try:
            for leaf in objs:
                gfile = leaf.get_gfile()
                gfile.trash()

        except GLib.Error as exc:
            # pylint: disable=no-member
            raise OperationError(exc.message) from exc

    def valid_for_item(self, leaf):
        gfile = leaf.get_gfile()
        if not gfile.query_exists(None):
            return False

        info = gfile.query_info(
            Gio.FILE_ATTRIBUTE_ACCESS_CAN_TRASH,
            Gio.FileQueryInfoFlags.NONE,
            None,
        )
        return info.get_attribute_boolean(Gio.FILE_ATTRIBUTE_ACCESS_CAN_TRASH)

    def get_description(self):
        return _("Move this file to trash")

    def get_icon_name(self):
        return "user-trash-full"

    def item_types(self):
        yield FileLeaf


class RestoreTrashedFile(Action):
    def __init__(self):
        Action.__init__(self, _("Restore"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        orig_path = leaf.get_orig_path()
        if not orig_path:
            return None

        orig_gfile = Gio.File.new_for_path(orig_path)
        cur_gfile = leaf.get_gfile()
        if orig_gfile.query_exists():
            raise OperationError(
                f"Target file exists at {orig_gfile.get_path()}"
            )

        pretty.print_debug(__name__, f"Move {cur_gfile} to {orig_gfile}")
        ret = cur_gfile.move(
            orig_gfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None
        )
        pretty.print_debug(__name__, f"Move ret={ret}")
        return FileLeaf(orig_gfile.get_path())

    def get_description(self):
        return _("Move file back to original location")

    def get_icon_name(self):
        return "edit-undo"


class EmptyTrash(Action):
    rank_adjust = -1

    def __init__(self):
        Action.__init__(self, _("Empty Trash"))

    def activate(self, leaf, iobj=None, ctx=None):
        gfile = Gio.File.new_for_uri(_TRASH_URI)
        failed = []
        for info in gfile.enumerate_children(
            "standard::*,trash::*", Gio.FileQueryInfoFlags.NONE, None
        ):
            name = info.get_name()
            if not gfile.get_child(name).delete():
                failed.append(name)

        if failed:
            err = _("Could not delete files:\n    ")
            raise OperationError(err + "\n    ".join(failed))

    def get_icon_name(self):
        return "user-trash-full"


class TrashFile(Leaf):
    """A file in the trash. Represented object is a file info object"""

    def __init__(self, trash_uri, info):
        name = info.get_display_name()
        Leaf.__init__(self, info, name)
        self._trash_uri = trash_uri

    def get_actions(self):
        if self.get_orig_path():
            yield RestoreTrashedFile()

    def get_gfile(self):
        return Gio.File.new_for_uri(self._trash_uri).get_child(
            self.object.get_name()
        )

    def get_orig_path(self):
        with suppress(AttributeError):
            return self.object.get_attribute_byte_string("trash::orig-path")

        return None

    def is_valid(self):
        return self.get_gfile().query_exists()

    def get_description(self):
        orig_path = self.get_orig_path()
        return (
            launch.get_display_path_for_bytestring(orig_path)
            if orig_path
            else None
        )

    def get_gicon(self):
        return self.object.get_icon()

    def get_icon_name(self):
        return "text-x-generic"


class TrashContentSource(Source):
    def __init__(self, trash_uri, name):
        Source.__init__(self, name)
        self._trash_uri = trash_uri

    def is_dynamic(self):
        return True

    def get_items(self):
        gfile = Gio.File.new_for_uri(self._trash_uri)
        for info in gfile.enumerate_children(
            "standard::*,trash::*", Gio.FileQueryInfoFlags.NONE, None
        ):
            yield TrashFile(self._trash_uri, info)

    def should_sort_lexically(self):
        return True

    def get_gicon(self):
        return icons.get_gicon_for_file(self._trash_uri)


class SpecialLocation(Leaf):
    """Base class for Special locations (in GIO/GVFS),
    such as trash:/// Here we assume they are all "directories"
    """

    def __init__(self, location, name=None, description=None, icon_name=None):
        """Special location with @location and
        @name. If unset, we find @name from filesystem
        @description is Leaf description"""
        gfile = Gio.File.new_for_uri(location)
        info = gfile.query_info(
            Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
            Gio.FileQueryInfoFlags.NONE,
            None,
        )
        name = (
            info.get_attribute_string(Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME)
            or location
        )
        Leaf.__init__(self, location, name)
        self.description = description
        self.icon_name = icon_name

    def get_actions(self):
        yield OpenTrash()

    def get_description(self):
        return self.description or self.object

    def get_gicon(self):
        # Get icon
        return icons.get_gicon_for_file(self.object)

    def get_icon_name(self):
        return "folder"


class Trash(SpecialLocation):
    def __init__(self, trash_uri, name=None):
        SpecialLocation.__init__(self, trash_uri, name=name)

    def has_content(self):
        return self.get_item_count()

    def content_source(self, alternate=False):
        return TrashContentSource(self.object, name=str(self))

    def get_actions(self):
        yield from SpecialLocation.get_actions(self)
        if self.get_item_count():
            yield EmptyTrash()

    def get_item_count(self):
        gfile = Gio.File.new_for_uri(self.object)
        info = gfile.query_info(
            Gio.FILE_ATTRIBUTE_TRASH_ITEM_COUNT,
            Gio.FileQueryInfoFlags.NONE,
            None,
        )
        return info.get_attribute_uint32(Gio.FILE_ATTRIBUTE_TRASH_ITEM_COUNT)

    def get_description(self):
        item_count = self.get_item_count()
        if not item_count:
            return _("Trash is empty")

        # proper translation of plural
        return ngettext(
            "Trash contains one file",
            "Trash contains %(num)s files",
            item_count,
        ) % {"num": item_count}


class InvisibleSourceLeaf(SourceLeaf):
    """Hack to hide this source"""

    def is_valid(self):
        return False


class TrashSource(Source):
    def __init__(self):
        Source.__init__(self, _("Trash"))

    def get_items(self):
        try:
            yield Trash(_TRASH_URI)
        except GLib.Error:
            self.output_exc()

    def get_leaf_repr(self):
        return InvisibleSourceLeaf(self)

    def provides(self):
        yield SpecialLocation

    def get_icon_name(self):
        return "user-trash"


class OpenTrash(Open):
    def activate(self, leaf, iobj=None, ctx=None):
        launch.show_url(leaf.object)



from __future__ import annotations

import typing as ty
import urllib.parse
from contextlib import suppress
from urllib.parse import urlparse as _urlparse
from urllib.parse import urlunparse as _urlunparse

from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from kupfer.obj.base import Leaf, Source

__all__ = ("QFURL_SCHEME", "Qfurl", "QfurlError")

QFURL_SCHEME = "qpfer"

# One would hope that there was a better way to do this
urllib.parse.uses_netloc.append(QFURL_SCHEME)
with suppress(AttributeError):
    urllib.parse.uses_fragment.append(QFURL_SCHEME)


class QfurlError(Exception):
    pass


class Qfurl:
    """A Qfurl is a URI to locate unique objects in kupfer's catalog.

    The Qfurl is built up as follows:
    ``qpfer://mother/qfid#module_and_type_hint``

    The mother part is a mother source identifier and is optional.
    The module_and_type_hint is optional.

    A short url looks like the following:
    ``qpfer:identifier``

    This class provides methods to get the Qfurl for an object,
    and resolve the object in a catalog.

    >>> class Object (object):
    ...     qf_id = "token"
    ...
    >>> q = Qfurl(Object())
    >>> Qfurl.reduce_url(q.url)
    'qpfer:token'

    >>> class Source (object):
    ...     def get_leaves(self):
    ...         yield Object()
    ...     def provides(self):
    ...         yield Object
    ...
    >>> q.resolve_in_catalog((Source(), ))  # doctest: +ELLIPSIS
    <__main__.Object object at 0x...>
    """

    def __init__(self, obj=None, url=None):
        """Create a new Qfurl for object @obj"""
        if obj:
            typname = f"{type(obj).__module__}.{type(obj).__name__}"
            try:
                qfid = obj.qf_id
            except AttributeError as exe:
                raise QfurlError(f"{obj} has no Qfurl") from exe

            self.url = _urlunparse((QFURL_SCHEME, "", qfid, "", "", typname))
        else:
            self.url = url

    def __str__(self) -> str:
        return self.url

    def __hash__(self) -> int:
        return hash(self.url)

    def __eq__(self, other: object) -> bool:
        return self.reduce_url(self.url) == self.reduce_url(other.url)  # type: ignore

    @classmethod
    def reduce_url(cls, url: str) -> str:
        """
        >>> url = "qpfer://mother/qfid#module_and_type_hint"
        >>> Qfurl.reduce_url(url)
        'qpfer://mother/qfid'
        """
        return urllib.parse.urldefrag(url)[0].replace("///", "", 1)

    @classmethod
    def _parts_mother_id_typename(cls, url: str) -> tuple[str, str, str]:
        """
        >>> murl = "qpfer://mother/qfid#module_and_type_hint"
        >>> Qfurl._parts_mother_id_typename(murl)
        ('mother', 'qfid', 'module_and_type_hint')
        """
        scheme, mother, qfid, _ign, _ore, typname = _urlparse(url)
        if scheme != QFURL_SCHEME:
            raise QfurlError(f"Wrong scheme: {scheme}")

        qfid = qfid.lstrip("/")
        return mother, qfid, typname

    def resolve_in_catalog(
        self, catalog: ty.Collection[Source]
    ) -> Leaf | None:
        """Resolve self in a catalog of sources.
        Return *immediately* on match found"""
        _mother, _qfid, typname = self._parts_mother_id_typename(self.url)
        _module, name = typname.rsplit(".", 1) if typname else (None, None)
        for src in catalog:
            if (
                name
                and name not in (pt.__name__ for pt in src.provides())
                and name
                not in (
                    t.__name__
                    for pt in src.provides()
                    for t in pt.__subclasses__()
                )
            ):
                continue

            for obj in src.get_leaves() or []:
                if not hasattr(obj, "qf_id"):
                    continue

                with suppress(QfurlError):
                    if self == Qfurl(obj):
                        return obj

        pretty.print_debug(__name__, "No match found for", self)
        return None


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
virtualbox_vboxapi_support.py

Control VirtualBox via Python interface (vboxapi).
Only (?) Sun VirtualBox (no OSE).
"""

__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = "2018-10-21"

import vboxapi  # pylint: disable=import-error

from kupfer.plugin.virtualbox import constants as vbox_const
from kupfer.support import pretty

# check api
try:
    _x = vboxapi.VirtualBoxReflectionInfo(None).SessionState_Locked
except AttributeError as exc:
    raise ImportError from exc


MONITORED_DIRS = None
IS_DYNAMIC = False
ICON = "VBox"
APP_ID = "virtualbox"


_ACTIONS = {
    vbox_const.VM_POWEROFF: lambda c: c.powerDown(),
    vbox_const.VM_ACPI_POWEROFF: lambda c: c.powerButton(),
    vbox_const.VM_PAUSE: lambda c: c.pause(),
    vbox_const.VM_REBOOT: lambda c: c.reset(),
    vbox_const.VM_RESUME: lambda c: c.resume(),
    vbox_const.VM_SAVE: lambda c: c.saveState(),
}


_VBOX = vboxapi.VirtualBoxManager(None, None)


def get_machine_by_id(mid):
    try:
        mach = _VBOX.getVirtualBox().getMachine(mid)
    except AttributeError:
        mach = _VBOX.getVirtualBox().findMachine(mid)

    return mach


def get_machine_state(machine_id):
    """check vms state (on/off/paused)"""
    if _VBOX is None:
        return vbox_const.VM_STATE_POWEROFF

    state = vbox_const.VM_STATE_POWERON
    machine_state = None
    try:
        machine = get_machine_by_id(machine_id)
        pretty.print_debug(__name__, "machine ", machine)
        machine_state = machine.state
        if machine_state == _VBOX.constants.MachineState_Paused:
            state = vbox_const.VM_STATE_PAUSED
        elif machine_state in (
            _VBOX.constants.MachineState_PoweredOff,
            _VBOX.constants.MachineState_Aborted,
            _VBOX.constants.MachineState_Starting,
        ):
            state = vbox_const.VM_STATE_POWEROFF
        elif machine_state == _VBOX.constants.MachineState_Saved:
            state = vbox_const.VM_STATE_SAVED

    except Exception as err:  # exception == machine is off (xpcom.Exception)
        pretty.print_debug(__name__, "get_machine_state", machine_state, err)
        # silently set state to off
        state = vbox_const.VM_STATE_POWEROFF

    return state


def _machine_start(vm_uuid, mode):
    """Start virtual machine
    @param vm_uuid - uuid of virtual machine
    @param mode - mode: gui, headless
    """
    try:
        session = _VBOX.getSessionObject()
        mach = get_machine_by_id(vm_uuid)
        remote_sess = mach.launchVMProcess(session, mode, "")
        remote_sess.waitForCompletion(-1)
        session.unlockMachine()
    except Exception as err:
        pretty.print_error(
            __name__, "StartVM:", vm_uuid, "Mode ", mode, "error", err
        )


def _execute_machine_action(vm_uuid, action):
    """Start virtual machine
    @param vm_uuid - uuid of virtual machine
    @param action - function called on _VBOX session
    """
    try:
        session = _VBOX.getSessionObject()
        mach = get_machine_by_id(vm_uuid)
        mach.lockMachine(session, _VBOX.constants.LockType_Shared)
        action(session.console)
        session.unlockMachine()
    except Exception as err:
        pretty.print_error(
            __name__,
            "_execute_machine_action:",
            action,
            " vm:",
            vm_uuid,
            "error",
            err,
        )


def vm_action(action, vm_uuid):
    """change state of the virtual machine
    @param action - one of the const VM_*
    @param vm_uuid - virtual machine uuid
    """
    if _VBOX is None:
        return

    if action == vbox_const.VM_START_NORMAL:
        _machine_start(vm_uuid, "gui")
    elif action == vbox_const.VM_START_HEADLESS:
        _machine_start(vm_uuid, "headless")
    else:
        command = _ACTIONS[action]
        _execute_machine_action(vm_uuid, command)


def get_machines():
    """Get generator of items:
    (machine uuid, machine name, machine description)
    """
    if _VBOX is None:
        return

    machines = _VBOX.getArray(_VBOX.getVirtualBox(), "machines")
    for machine in machines:
        if machine.accessible:
            description = machine.description or machine.OSTypeId
            yield (machine.id, machine.name, description)


def unload():
    pass



__kupfer_name__ = _("Firefox Tags")
__kupfer_sources__ = ("TagsSource",)
__description__ = _("Browse Firefox bookmarks by tags")
__version__ = "2021-09-05"
__author__ = "Karol Będkowski"

import itertools
import typing as ty

from kupfer import plugin_support
from kupfer.obj import Leaf, Source, UrlLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.plugin._firefox_support import (
    get_firefox_home_file,
    query_database,
)

if ty.TYPE_CHECKING:
    from gettext import gettext as _


MAX_ITEMS = 10000

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "profile",
        "label": _("Firefox profile name or path"),
        "type": str,
        "value": "",
        "helper": "choose_directory",
    },
)


class FirefoxTag(Leaf):
    def get_icon_name(self):
        return "text-html"

    def has_content(self):
        return True

    def content_source(self, alternate=False):
        return TagBookmarksSource(self.object, self.name)


_TAGS_SQL = """
SELECT id, title
FROM moz_bookmarks mb
WHERE mb.parent = 4
    AND mb.fk IS NULL
"""


class TagsSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("firefox", "firefox-esr")
    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Firefox Tags"))
        self.monitor_token = None

    def initialize(self):
        profile = __kupfer_settings__["profile"]
        if ff_home := get_firefox_home_file("", profile):
            self.monitor_token = self.monitor_directories(str(ff_home))

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == "lock"

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def get_items(self):
        """Get tags from firefox places database"""
        fpath = get_firefox_home_file(
            "places.sqlite", __kupfer_settings__["profile"]
        )
        if not fpath:
            return []

        return list(
            itertools.starmap(
                FirefoxTag, query_database(str(fpath), _TAGS_SQL)
            )
        )

    def get_description(self):
        return _("Index of Firefox bookmarks by tags")

    def get_gicon(self):
        if lrepr := self.get_leaf_repr():
            return lrepr.get_gicon()

        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield FirefoxTag


_TAG_BOOKMARKS_SQL = """
SELECT mp.url, mp.title
FROM moz_places mp
JOIN moz_bookmarks mb ON mp.id = mb.fk
WHERE mb.keyword_id IS NULL
	AND mb.parent = ?
ORDER BY visit_count DESC
LIMIT ?"""


class TagBookmarksSource(Source):
    def __init__(self, tag_id: int, tag: str):
        super().__init__(_("Firefox Bookmarks by tag"))
        self.tag = tag
        self.tag_id = tag_id

    def get_items(self):
        """Query the firefox places database for bookmarks with tag"""
        fpath = get_firefox_home_file(
            "places.sqlite", __kupfer_settings__["profile"]
        )
        if not fpath:
            return []

        return list(
            itertools.starmap(
                UrlLeaf,
                query_database(
                    str(fpath), _TAG_BOOKMARKS_SQL, (self.tag_id, MAX_ITEMS)
                ),
            )
        )

    def get_gicon(self):
        if lrepr := self.get_leaf_repr():
            return lrepr.get_gicon()

        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield UrlLeaf



"""
Copyright 2007--2009 Ulrik Sverdrup <ulrik.sverdrup@gmail.com>

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty
import zlib

from kupfer import icons
from kupfer.obj import actions
from kupfer.obj.base import Action, Leaf, Source
from kupfer.obj.representation import TextRepresentation

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext

    from kupfer.core import commandexec


__all__ = (
    "RunnableLeaf",
    "SourceLeaf",
    "TextLeaf",
    "UrlLeaf",
)


class SourceLeaf(Leaf):
    def __init__(self, obj: Source, name: str | None = None) -> None:
        """Create SourceLeaf for source @obj.
        Represented object is source."""
        Leaf.__init__(self, obj, name or str(obj))

    def has_content(self) -> bool:
        return True

    def repr_key(self) -> str:
        return repr(self.object)

    def content_source(self, alternate: bool = False) -> Source:
        assert isinstance(self.object, Source)
        return self.object

    def get_description(self) -> str | None:
        assert isinstance(self.object, Source)
        return self.object.get_description()

    @property
    def fallback_icon_name(self) -> str:  # type: ignore
        assert isinstance(self.object, Source)
        return self.object.fallback_icon_name

    def get_gicon(self) -> icons.GIcon | None:
        return self.object.get_gicon()

    def get_icon_name(self) -> str:
        assert isinstance(self.object, Source)
        return self.object.get_icon_name()


class UrlLeaf(Leaf, TextRepresentation):
    serializable = 1

    def __init__(self, obj: str, name: str | None) -> None:
        super().__init__(obj, name or obj)
        if obj != name:
            self.kupfer_add_alias(obj)

    def get_actions(self) -> ty.Iterator[Action]:
        yield actions.OpenUrl()

    def get_description(self) -> str:
        return self.object  # type:ignore

    def get_icon_name(self) -> str:
        return "text-html"


class RunnableLeaf(Leaf):
    """Leaf where the Leaf is basically the action itself, for items such
    as Quit, Log out etc."""

    def __init__(self, obj: ty.Any = None, name: str = "") -> None:
        Leaf.__init__(self, obj, name)

    def get_actions(self) -> ty.Iterator[Action]:
        yield actions.Perform(
            has_result=self.has_result(), item_types=tuple(self.item_types())
        )

    def has_result(self) -> bool:
        return False

    def item_types(self) -> ty.Iterable[Leaf]:
        return ()

    def run(self, ctx: commandexec.ExecutionToken | None = None) -> ty.Any:
        raise NotImplementedError

    def wants_context(self) -> bool:
        """Return ``True`` if you want the actions' execution
        context passed as ctx= in RunnableLeaf.run."""
        return False

    def repr_key(self) -> ty.Any:
        return None

    def get_gicon(self) -> icons.GIcon | None:
        if iname := self.get_icon_name():
            return icons.get_gicon_with_fallbacks(None, (iname,))

        return icons.ComposedIcon("kupfer-object", "kupfer-execute")

    def get_icon_name(self) -> str:
        return ""


class TextLeaf(Leaf, TextRepresentation):
    """Represent a text query.
    The represented object is a string"""

    serializable = 1

    def __init__(self, text: str, name: str | None = None) -> None:
        """@text *must* be unicode or UTF-8 str"""
        if not name:
            name = self.get_first_text_line(text)

        if not text or not name:
            name = _("(Empty Text)")

        assert name
        Leaf.__init__(self, text, name)

    def repr_key(self) -> ty.Any:
        return zlib.crc32(self.object.encode("utf-8", "surrogateescape"))

    @classmethod
    def get_first_text_line(cls, text: str) -> str:
        if not text:
            return text

        firstline, *_dummy = text.lstrip().partition("\n")
        return firstline

    def get_description(self) -> str:
        numlines = self.object.count("\n") + 1
        desc = self.get_first_text_line(self.object)

        # TRANS: This is description for a TextLeaf, a free-text search
        # TRANS: The plural parameter is the number of lines %(num)d
        return ngettext(
            '"%(text)s"', '(%(num)d lines) "%(text)s"', numlines
        ) % {"num": numlines, "text": desc}

    def get_icon_name(self) -> str:
        return "edit-select-all"


class _ExecuteAction(Action):
    """Execute action selected as leaf on iobj.
    This is special action used in "action -> object" mode.

    For most this action copy properties from real action; only replace
    leaves with indirect objects.
    """

    rank_adjust = 99999

    def __init__(self, action: Action) -> None:
        super().__init__(name=_("Run On..."))
        self.action = action

    def wants_context(self) -> bool:
        return self.action.wants_context()

    def has_result(self):
        return self.action.has_result()

    def is_factory(self):
        return self.action.is_factory()

    def is_async(self):
        return self.action.is_async()

    def activate(self, leaf, iobj=None, ctx=None):
        """For this action leaf is Action type, iobj is
        leaf of type accepted by action."""
        assert iobj
        return self.action.activate(iobj, ctx=ctx)

    def requires_object(self):
        return True

    def object_types(self):
        yield from self.action.item_types()

    def valid_object(self, iobj, for_item=None):
        return self.action.valid_for_item(iobj)

    def repr_key(self) -> ty.Any:
        return self.action


class ActionLeaf(Leaf):
    """ActionLeaf is special leaf that carry real "action". Is used in two-panes
    mode (Action -> object) and allow user to select action in first panel.
    """

    object: Action

    serializable = None

    def __init__(self, action: Action):
        Leaf.__init__(self, action, action.name)

    def get_actions(self):
        return (_ExecuteAction(self.object),)

    def get_gicon(self):
        return self.object.get_gicon()

    def get_icon_name(self):
        return self.object.get_icon_name()

    def repr_key(self) -> ty.Any:
        return self.object

    def get_description(self):
        return self.object.get_description()



from __future__ import annotations

__kupfer_name__ = _("Audacious")
__kupfer_sources__ = ("AudaciousSource",)
__description__ = _("Control Audacious playback and playlist")
__version__ = "2023.1"
__author__ = "Horia V. Corcalciuc <h.v.corcalciuc@gmail.com>, US, KB"

# TODO: support many playlists (?); for now we load only active playlist

import subprocess
import time
import typing as ty

import dbus

from kupfer import icons, launch, plugin_support
from kupfer.obj import (
    Action,
    Leaf,
    NotAvailableError,
    RunnableLeaf,
    Source,
    SourceLeaf,
)
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.support import kupferstring, pretty, weaklib
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "playlist_toplevel",
        "label": _("Include songs in top level"),
        "type": bool,
        "value": True,
    },
)


_AUDTOOL = "audtool"
_AUDACIOUS = "audacious"
_BUS_NAME = "org.atheme.audacious"
_OBJ_NAME = "/org/atheme/audacious"
_IFACE_NAME = "org.atheme.audacious"


plugin_support.check_command_available(_AUDTOOL)


def _create_dbus_connection(
    iface: str, obj: str, service: str
) -> dbus.Interface:
    """Create dbus connection to NetworkManager"""

    sbus = dbus.SessionBus()
    try:
        if dobj := sbus.get_object(service, obj):
            return dbus.Interface(dobj, iface)

    except dbus.exceptions.DBusException as err:
        pretty.print_debug(__name__, err)
        raise NotAvailableError(service) from err

    raise NotAvailableError(service)


def get_playlist_songs_dbus() -> ty.Iterator[SongLeaf]:
    conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
    total = conn.Length()
    for pos in range(1, total + 1):
        title = " ".join(
            (conn.SongTuple(pos, "title"), conn.SongTuple(pos, "artist"))
        ).strip()

        if not title:
            title = conn.SongTitle(pos)

        yield SongLeaf(pos, title)


def get_playlist_songs() -> ty.Iterator[SongLeaf]:
    """Yield tuples of (position, name) for playlist songs"""
    with subprocess.Popen(
        [_AUDTOOL, "playlist-display"], stdout=subprocess.PIPE
    ) as proc:
        stdout, _stderr = proc.communicate()
        for line in stdout.splitlines():
            if not line.count(b"|") >= 2:  # noqa:PLR2004
                continue

            position, rest = line.split(b"|", 1)
            songname, rest = rest.rsplit(b"|", 1)
            pos = int(position.strip())
            name = kupferstring.fromlocale(songname.strip())
            yield SongLeaf(pos, name)


def get_current_song() -> str | None:
    try:
        conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
        pos = conn.Position()
        total = conn.Length()
        return "\n".join(
            filter(
                None,
                (
                    conn.SongTuple(pos, "title"),
                    conn.SongTuple(pos, "artist"),
                    conn.SongTitle(pos),
                    _("Position: %(pos)d / %(total)d")
                    % {"pos": pos, "total": total},
                ),
            )
        )
    except Exception:
        pretty.print_exc(__name__)

    with subprocess.Popen(
        [_AUDTOOL, "current-song"], stdout=subprocess.PIPE
    ) as proc:
        stdout, _stderr = proc.communicate()
        for line in stdout.splitlines():
            return kupferstring.fromlocale(line)

        return None


def _audtool_command(args: tuple[str, ...], activate: bool = True) -> None:
    def _cb(res: int) -> None:
        if res == 0:
            return

        # audacious not running
        launch.spawn_async(("audacious",))
        time.sleep(1)
        launch.spawn_async(args)

    launch.spawn_async(args, finish_callback=_cb)


class Enqueue(Action):
    action_accelerator = "e"

    def __init__(self):
        Action.__init__(self, _("Enqueue"))

    def activate(self, leaf, iobj=None, ctx=None):
        pos = leaf.object
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.PlayqueneAdd(pos)
        except Exception:
            pretty.print_exc(__name__)
            launch.spawn_async((_AUDTOOL, "playqueue-add", str(pos)))

    def get_description(self):
        return _("Add track to the Audacious play queue")

    def get_gicon(self):
        return icons.ComposedIcon("gtk-execute", "media-playback-start")

    def get_icon_name(self):
        return "media-playback-start"


class Dequeue(Action):
    def __init__(self):
        Action.__init__(self, _("Dequeue"))

    def activate(self, leaf, iobj=None, ctx=None):
        pos = leaf.object
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.PlayqueneRemove(pos)
        except Exception:
            launch.spawn_async((_AUDTOOL, "playqueue-remove", str(pos)))

    def get_description(self):
        return _("Remove track from the Audacious play queue")

    def get_gicon(self):
        return icons.ComposedIcon("gtk-execute", "media-playback-stop")

    def get_icon_name(self):
        return "media-playback-stop"


class JumpToSong(Action):
    action_accelerator = "o"

    def __init__(self):
        Action.__init__(self, _("Play"))

    def activate(self, leaf, iobj=None, ctx=None):
        pos = leaf.object
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.Jump(pos)
            conn.Play()
        except Exception:
            _audtool_command((_AUDTOOL, "playlist-jump", str(pos)))
            time.sleep(1)
            launch.spawn_async((_AUDTOOL, "playback-play"))

    def get_description(self):
        return _("Jump to track in Audacious")

    def get_icon_name(self):
        return "media-playback-start"


class Play(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Play"))

    def run(self, ctx=None):
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.Play()
        except Exception:
            _audtool_command((_AUDTOOL, "playback-play"))

    def get_description(self):
        return _("Resume playback in Audacious")

    def get_icon_name(self):
        return "media-playback-start"


class Stop(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Stop"))

    def run(self, ctx=None):
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.Stop()
        except Exception:
            launch.spawn_async((_AUDTOOL, "playback-play"))

    def get_description(self):
        return _("Stop playback in Audacious")

    def get_icon_name(self):
        return "media-playback-stop"


class Pause(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Pause"))

    def run(self, ctx=None):
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.Pause()
        except Exception:
            launch.spawn_async((_AUDTOOL, "playback-pause"))

    def get_description(self):
        return _("Pause playback in Audacious")

    def get_icon_name(self):
        return "media-playback-pause"


class Next(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Next"))

    def run(self, ctx=None):
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.Advance()
        except Exception:
            _audtool_command((_AUDTOOL, "playlist-advance"))

    def get_description(self):
        return _("Jump to next track in Audacious")

    def get_icon_name(self):
        return "media-skip-forward"


class Previous(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Previous"))

    def run(self, ctx=None):
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.Reverse()
        except Exception:
            _audtool_command((_AUDTOOL, "playlist-reverse"))

    def get_description(self):
        return _("Jump to previous track in Audacious")

    def get_icon_name(self):
        return "media-skip-backward"


class ClearQueue(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Clear Queue"))

    def run(self, ctx=None):
        try:
            conn = _create_dbus_connection(_IFACE_NAME, _OBJ_NAME, _BUS_NAME)
            conn.PlayqueneClear()
        except Exception:
            launch.spawn_async((_AUDTOOL, "playqueue-clear"))

    def get_description(self):
        return _("Clear the Audacious play queue")

    def get_icon_name(self):
        return "edit-clear"


class Shuffle(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Shuffle"))

    def run(self, ctx=None):
        launch.spawn_async((_AUDTOOL, "playlist-shuffle-toggle"))

    def get_description(self):
        return _("Toggle shuffle in Audacious")

    def get_icon_name(self):
        return "media-playlist-shuffle"


class Repeat(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Repeat"))

    def run(self, ctx=None):
        launch.spawn_async((_AUDTOOL, "playlist-repeat-toggle"))

    def get_description(self):
        return _("Toggle repeat in Audacious")

    def get_icon_name(self):
        return "media-playlist-repeat"


class ShowPlaying(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Show Playing"))

    def run(self, ctx=None):
        song = get_current_song()
        if song is not None:
            uiutils.show_notification(song, icon_name="audio-x-generic")

    def get_description(self):
        return _("Tell which song is currently playing")

    def get_gicon(self):
        return icons.ComposedIcon("dialog-information", "audio-x-generic")

    def get_icon_name(self):
        return "dialog-information"


class SongLeaf(Leaf):
    """The SongLeaf's represented object is the Playlist index"""

    def get_actions(self):
        yield JumpToSong()
        yield Enqueue()
        yield Dequeue()

    def get_icon_name(self):
        return "audio-x-generic"

    def get_text_representation(self):
        return self.name


class AudaciousSongsSource(Source):
    def __init__(self):
        Source.__init__(self, _("Playlist"))

    def get_items(self):
        try:
            return list(get_playlist_songs_dbus())
        except Exception:
            pretty.print_exc(__name__)
            return list(get_playlist_songs())

    def get_gicon(self):
        return icons.ComposedIcon(
            _AUDACIOUS, "audio-x-generic", emblem_is_fallback=True
        )

    def provides(self):
        yield SongLeaf


class AudaciousSource(AppLeafContentMixin, Source):
    appleaf_content_id = _AUDACIOUS
    source_user_reloadable = True
    source_scan_interval: int = 3600

    def __init__(self):
        Source.__init__(self, _("Audacious"))

    def initialize(self):
        bus = dbus.SessionBus()
        weaklib.dbus_signal_connect_weakly(
            bus,
            "NameOwnerChanged",
            self._name_owner_changed,
            dbus_interface="org.freedesktop.DBus",
            arg0=_BUS_NAME,
        )

    def _name_owner_changed(self, name, old, new):
        if new:
            self.mark_for_update(postpone=True)

    def get_items(self):
        yield Play()
        yield Stop()
        yield Pause()
        yield Next()
        yield Previous()
        yield ClearQueue()
        yield ShowPlaying()
        # Commented as these seem to have no effect
        # yield Shuffle()
        # yield Repeat()
        songs_source = AudaciousSongsSource()
        yield SourceLeaf(songs_source)
        if __kupfer_settings__["playlist_toplevel"]:
            yield from songs_source.get_leaves()

    def get_description(self):
        return __description__

    def get_icon_name(self):
        return _AUDACIOUS

    def provides(self):
        yield RunnableLeaf



"""
Changes:
    2018-09-04 * fix: Python 3 compatibility
    2012-10-08 * fix: errors when no one configured directories exists
    2012-06-09 + max depth; fix source name
    2012-06-08 - init
"""

__kupfer_name__ = _("Deep Directories")
__kupfer_sources__ = ("DeepDirSource",)
__description__ = _("Recursive index directories")
__version__ = "2018-09-04"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import os
import typing as ty

from kupfer import plugin_support
from kupfer.obj.filesrc import FileSource

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "dirs",
        "label": _("Directories:"),
        "type": list,
        "helper": "choose_directory",
        "value": ["~/Documents/"],
    },
    {
        "key": "depth",
        "label": _("Depth (max 10):"),
        "type": int,
        "value": 2,
        "max": 10,
        "min": 1,
    },
)


_MAX_DEPTH = 10


class DeepDirSource(FileSource):
    def __init__(self, name=_("Deep Directories")):
        FileSource.__init__(
            self,
            [""],
            min(__kupfer_settings__["depth"], _MAX_DEPTH),
        )
        self.name = name

    def initialize(self):
        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )

    def get_items(self):
        self.dirlist = list(self._get_dirs())
        self.depth = min(__kupfer_settings__["depth"], _MAX_DEPTH)
        yield from FileSource.get_items(self)

    @staticmethod
    def _get_dirs():
        dirs = __kupfer_settings__["dirs"]
        for path in dirs or ():
            if (path := os.path.expanduser(path.strip())) and os.path.isdir(
                path
            ):
                yield path

    def _on_setting_changed(self, settings, key, value):
        if key in ("dirs", "depth"):
            self.mark_for_update()



__kupfer_name__ = "Core"
# The following attributes are updated later
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_contents__ = ()
__kupfer_actions__ = ("SearchInside", "CopyToClipboard", "Rescan")
__description__ = "Core actions and items"
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import typing as ty
from contextlib import suppress

from gi.repository import Gdk, Gtk

from kupfer import interface
from kupfer.core import commandexec
from kupfer.obj import Action, Leaf, Source, SourceLeaf
from kupfer.obj.apps import AppLeaf
from kupfer.obj.exceptions import InvalidLeafError
from kupfer.obj.sources import MultiSource
from kupfer.plugin.core import commands, contents, internal, text
from kupfer.support import pretty, task

if ty.TYPE_CHECKING:
    from gettext import gettext as _


def _is_debug():
    # Return True if Kupfer is in debug mode
    return pretty.DEBUG


def register_subplugin(module):
    attrs = (
        "__kupfer_sources__",
        "__kupfer_actions__",
        "__kupfer_text_sources__",
        "__kupfer_contents__",
    )
    for attr in attrs:
        object_names = getattr(module, attr, ())
        globals()[attr] += object_names
        globals().update((sym, getattr(module, sym)) for sym in object_names)


register_subplugin(contents)
register_subplugin(text)
register_subplugin(internal)
register_subplugin(commands)

if _is_debug():
    from kupfer.plugin.core import debug

    register_subplugin(debug)


def initialize_plugin(x):
    # pylint: disable=import-outside-toplevel
    from kupfer.plugin.core import alternatives

    alternatives.initialize_alternatives(__name__)


class _MultiSource(MultiSource):
    def is_dynamic(self):
        return False


class SearchInside(Action):
    """Return the content source for a Leaf"""

    def __init__(self):
        super().__init__(_("Search Contents"))

    def is_factory(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        if not leaf.has_content():
            raise InvalidLeafError("Must have content")

        return leaf.content_source()

    def activate_multiple(self, objs):
        return _MultiSource([L.content_source() for L in objs])

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        return leaf.has_content()

    def get_description(self):
        return _("Search inside this catalog")

    def get_icon_name(self):
        return "kupfer-search"


class CopyToClipboard(Action):
    # rank down since it applies everywhere
    rank_adjust = -2
    action_accelerator = "c"

    def __init__(self):
        Action.__init__(self, _("Copy"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        clip = Gtk.Clipboard.get_for_display(
            ctx.environment.get_screen().get_display(), Gdk.SELECTION_CLIPBOARD
        )
        interface.copy_to_clipboard(leaf, clip)

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        with suppress(AttributeError):
            return bool(interface.get_text_representation(leaf))

    def get_description(self):
        return _("Copy to clipboard")

    def get_icon_name(self):
        return "edit-copy"


class RescanActionTask(task.ThreadTask):
    def __init__(
        self,
        source: Source,
        async_token: commandexec.ExecutionToken,
        retval: Leaf,
    ) -> None:
        task.ThreadTask.__init__(self)
        self.source = source
        self.async_token = async_token
        self.retval = retval

    def thread_do(self) -> None:
        self.source.get_leaves(force_update=True)

    def thread_finish(self) -> None:
        self.async_token.register_late_result(self.retval)


class Rescan(Action):
    """A source action: Rescan a source!"""

    rank_adjust = -5

    def __init__(self):
        Action.__init__(self, _("Rescan"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        if not leaf.has_content():
            raise InvalidLeafError("Must have content")

        source = leaf.content_source()
        return RescanActionTask(source, ctx, leaf)

    def is_async(self):
        return True

    def get_description(self):
        return _("Force reindex of this source")

    def get_icon_name(self):
        return "view-refresh"

    def item_types(self):
        yield AppLeaf
        yield SourceLeaf

    def valid_for_item(self, leaf):
        if not leaf.has_content():
            return False

        if leaf.content_source().is_dynamic():
            return False

        return _is_debug() or leaf.content_source().source_user_reloadable



from __future__ import annotations

import re
import typing as ty
import urllib.parse
from contextlib import suppress

__all__ = ("is_url", "is_valid_email", "is_valid_file_path", "validate_netloc")


def _validate_port(port: str) -> bool:
    if not port:
        return False

    with suppress(ValueError):
        return 0 <= int(port) <= 65535  # noqa: PLR2004

    return False


def _is_ipv4(string: str) -> bool:
    if len(octets := string.split(".")) == 4:  # noqa: PLR2004
        with suppress(ValueError):
            return all(0 <= int(o) <= 255 for o in octets)  # noqa: PLR2004

    return False


def _is_ipv6(netloc: str) -> bool:
    # quick and not perfect ipv6 address validation
    if netloc[0] != "[" or netloc[-1] != "]":
        return False

    netloc = netloc[1:-1]
    if not (2 <= len(netloc) <= 39):  # noqa: PLR2004
        return False

    parts = netloc.split(":")
    if not (3 <= len(parts) <= 8):  # noqa: PLR2004
        return False

    return all(re.match(r"[0-9a-fA-F]{1,4}", part) for part in parts if part)


def _is_valid_domain(domain: str) -> bool:
    if not domain:
        return False

    domain_re = (
        # hostname
        r"[a-z\u00a1-\uffff0-9](?:[a-z\u00a1-\uffff0-9-]{0,61}"
        r"[a-z\u00a1-\uffff0-9])?"
        # domain
        r"(?:\.(?!-)[a-z\u00a1-\uffff0-9-]{1,63}(?<!-))*"
        # tld
        r"\."  # dot
        r"(?!-)"  # can't start with a dash
        r"(?:[a-z\u00a1-\uffff-]{2,63}"  # domain label
        r"|xn--[a-z0-9]{1,59})"  # or punycode label
        r"(?<!-)"  # can't end with a dash
        r"\.?"  # may have a trailing dot
    )
    return bool(re.match(domain_re, domain, re.IGNORECASE))


def _is_valid_hostname(name: str) -> bool:
    return bool(
        re.match(
            r"[a-z\u00a1-\uffff0-9](?:[a-z\u00a1-\uffff0-9-]{0,61}"
            r"[a-z\u00a1-\uffff0-9])?",
            name,
            re.IGNORECASE,
        )
    )


def validate_netloc(netloc: str) -> bool:
    """Check if `netloc` is valid netlocation.

    Accepted forms [<user>[:<pass>]@]<hostname with domain|ipv4|ipv6>[:port]

    based on django.validator
    https://github.com/django/django/blob/main/django/core/validators.py
    """
    if not netloc:
        return False

    # netlock may contain user:pass
    if "@" in netloc:
        user_pass, _s, netloc = netloc.partition("@")
        if user_pass and not re.match(r"[^\s:@/]+(?::[^\s:@/]*)?", user_pass):
            return False

    host = netloc
    # is there domain:port?
    if re.match(r".+:\d{1,5}$", netloc):
        # check is port valid
        host, _s, port = netloc.rpartition(":")
        if not _validate_port(port):
            return False

    if not host:
        return False

    return (
        host == "localhost"
        or _is_ipv4(host)
        or _is_valid_domain(host)
        or _is_ipv6(host)
    )


def _guess_schema(netloc: str) -> str:
    """Guess schema for network location `netloc`."""
    if netloc.startswith("ftp.") or "@ftp." in netloc:
        return "ftp"

    # assume localhost is http
    if "localhost" in netloc or re.match(r"^127\.\d+\.\d+\.\d+$", netloc):
        return "http"

    return "https"


def _is_http_domain(netloc: str) -> bool:  # noqa: PLR0911
    """Guess is `netloc` is some kind of real domain. It's not accurate"""

    if netloc == "localhost":
        return True

    parts = netloc.split(".")
    if not all(parts):
        return False

    len_parts = len(parts)
    if len_parts == 1:
        return False

    if len_parts > 2:  # noqa: PLR2004
        return True

    if "localhost" in parts:
        return True

    lpart = parts[-1]
    if len(lpart) == 2:  # natonal code (mostly) # noqa: PLR2004
        return True

    return lpart in ("com", "org", "net", "gov", "local")


# from https://www.iana.org/assignments/uri-schemes/uri-schemes.xhtml
# schemes that not require / not have netloc. Skipped exotic/historic.
_schema_wo_netloc = {
    "android",
    "admin",  # unofficial, gnome
    "apt",
    "bitcoin",
    "bitcoincash",
    "callto",
    "dab",
    "dvb",
    "ethereum",
    "file",
    "fm",
    "gg",
    "gtalk",
    "ipfs",
    "ipns",
    "jabber",
    "lastfm",
    "magnet",
    "mailto",
    "maps",
    "market",
    "message",
    "nntp",
    "news",
    "skype",
    "sms",
    "steam",
}


def is_url(text: str) -> str | None:  # noqa: PLR0911
    """If `text` is an URL, return a cleaned-up URL, else return `None`

    Ref:
        https://en.wikipedia.org/wiki/Uniform_Resource_Identifier
        https://www.iana.org/assignments/uri-schemes/uri-schemes.xhtml
        https://en.wikipedia.org/wiki/List_of_URI_schemes
    """
    text = text.strip()

    # quick check is text contain valid for url characters
    if not re.match(
        r"^[a-z\u00a1-\uffff0-9-:/?#\[\]@!$&;()*+.,;=~_%-]+$",
        text,
        re.IGNORECASE,
    ):
        return None

    url = urllib.parse.urlparse(text)
    netloc = url.netloc
    scheme = url.scheme
    path = url.path

    if scheme and netloc:
        # if we have schema and valid domain - it's ok
        if validate_netloc(netloc):
            return text

        # accept also non-fqdn hostname
        if _is_valid_hostname(netloc):
            return text

        return None

    if scheme and path:
        # this is valid for mailto:, news: and other urls that not have netloc
        if scheme in _schema_wo_netloc:
            return text

        # spacial case for mail: replace mail: -> mailto:
        if scheme == "mail":
            return f"mailto{text[4:]}"

    if not netloc:
        # if there is no schema and no '//' on begin of netloc urlparse
        # put this in path value. Also when for address like 'abc:a@kks' 'abc'
        # is in schema and rest in path.
        # try to fix this:
        if scheme and "://" not in text:
            netloc, _d, path = text.partition("/")
            schema = ""
        elif path:
            netloc, _d, path = path.partition("/")

    # valid netloc
    if (
        netloc
        and validate_netloc(netloc)
        and (path or url.query or _is_http_domain(netloc))
    ):
        schema = _guess_schema(netloc)
        return f"{schema}://{text}"

    return None


def is_valid_email(text: str) -> bool:
    """Check if `text` is valid email address."""
    if not text:
        return False

    local_part, _dummy, domain = text.partition("@")
    if not _is_valid_domain(domain):
        return False

    re_local_part = r"[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+"
    return bool(re.match(re_local_part, local_part))


_UNPRINTABLE: ty.Final[str] = (
    "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0e\x0f"
    "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a"
    "\x1b\x1c\x1d\x1e\x1f\x7f"
)
_MAX_PATH_LEN: ty.Final[int] = 256


def is_valid_file_path(path: str | None) -> bool:
    """Not perfect path validation"""
    if not path:
        return False

    if len(path) > _MAX_PATH_LEN:
        return False

    if re.match(r"^[a-z]+://", path) or path.startswith("//"):
        return False

    return not any(c in _UNPRINTABLE for c in path)



from __future__ import annotations

__kupfer_name__ = _("XFCE Session Management")
__kupfer_sources__ = ("XfceItemsSource", "XfceWhskerFavoritesSource")
__description__ = _("Special items and actions for XFCE environment")
__version__ = "2023-05-17"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import typing as ty
from pathlib import Path

from kupfer import plugin_support, puid
from kupfer.core import learn
from kupfer.obj import Leaf, Source
from kupfer.plugin import session_support as support
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "lock_cmd",
        "label": _("Screen lock command"),
        "type": str,
        "value": "xflock4",
    },
    {
        "key": "whisker_favs",
        "label": _("Load Whisker Menu Favorite applications"),
        "type": bool,
        "value": True,
    },
)

# sequences of argument lists
_LOGOUT_CMD = (["xfce4-session-logout", "--logout"],)
_SHUTDOWN_CMD = (["xfce4-session-logout"],)


class XfceItemsSource(support.CommonSource):
    source_scan_interval: int = 36000

    def __init__(self):
        support.CommonSource.__init__(self, _("XFCE Session Management"))

    def get_items(self):
        lockscreen_cmd = (
            __kupfer_settings__["lock_cmd"] or "xdg-screensaver lock"
        )

        return (
            support.Logout(_LOGOUT_CMD),
            support.LockScreen((lockscreen_cmd.split(" "),)),
            support.Shutdown(_SHUTDOWN_CMD),
        )


def _find_whisher_conf_file() -> Path | None:
    base = Path("~/.config/xfce4/panel/").expanduser()
    files = (f for f in base.glob("whiskermenu*.rc") if f.is_file())
    sorted_files = sorted(files, key=lambda f: f.stat().st_mtime, reverse=True)
    return sorted_files[0] if sorted_files else None


def _get_whisker_conf_favs() -> list[str]:
    # find most recent configuration file
    conf_file = _find_whisher_conf_file()
    if not conf_file:
        return []

    pretty.print_debug(__name__, "loading whisker rc file", conf_file)

    with open(conf_file, encoding="UTF-8") as conf:
        for line in conf:
            if not line.startswith("favorites="):
                continue

            *_pref, favs = line.strip().partition("=")
            return favs.split(",")

    return []


def _load_whisher_favs() -> ty.Iterator[Leaf]:
    # simplified version
    for fav in _get_whisker_conf_favs():
        name, *_rest = fav.rpartition(".")
        id_ = f"<kupfer.obj.apps.AppLeaf {name}>"
        # ignore invalid objects
        if (
            (itm := puid.resolve_unique_id(id_))
            and (not hasattr(itm, "is_valid") or itm.is_valid())
            and isinstance(itm, Leaf)
        ):
            yield itm


class XfceWhskerFavoritesSource(Source):
    source_scan_interval: int = 1800
    serializable = None

    def __init__(self):
        Source.__init__(self, _("Whisker Favorites"))

    def initialize(self):
        self.mark_for_update()

    def finalize(self) -> None:
        learn.replace_favorites(__name__)

    def get_items(self):
        if not __kupfer_settings__["whisker_favs"]:
            return []

        try:
            favs = list(_load_whisher_favs())
        except OSError:
            favs = []
        else:
            learn.replace_favorites(__name__, *favs)

        return favs

    def get_description(self):
        return _("Favorite items from Xfce Whisker Menu")

    def get_icon_name(self):
        return "emblem-favorite"

    def is_valid(self):
        # for future use
        return __kupfer_settings__["whisker_favs"]



from contextlib import suppress

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
with suppress(ValueError):
    gi.require_version("Wnck", "3.0")

with suppress(ValueError):
    gi.require_version("AyatanaAppIndicator3", "0.1")
